import pickle
import sys

import emcee
from schwimmbad import MPIPool  # todo: add these as dependencies...?

from scope.run_simulation import *

do_pca = True
np.random.seed(42)

# load the data
test_data_path = os.path.join(os.path.dirname(__file__), "../data")


def log_prob(
    x,
    best_kp,
    wl_cube_model,
    Fp_conv,
    n_order,
    n_exposure,
    n_pixel,
    A_noplanet,
    star,
    n_princ_comp,
    flux_cube,
    wl_model,
    Fstar_conv,
    Rp_solar,
    Rstar,
    phases,
    do_pca,
):
    """
    just add the log likelihood and the log prob.

    Inputs
    ------
        :x: (array) array of parameters

    Outputs
    -------
        :log_prob: (float) log probability.
    """
    rv_semiamp_orbit = 0.0
    Kp, Vsys, log_scale = x
    scale = np.power(10, log_scale)
    prior_val = prior(x, best_kp)

    if not np.isfinite(prior_val):
        return -np.inf
    ll = calc_log_likelihood(
        Vsys,
        Kp,
        scale,
        wl_cube_model,
        wl_model,
        Fp_conv,
        Fstar_conv,
        flux_cube,
        n_order,
        n_exposure,
        n_pixel,
        phases,
        Rp_solar,
        Rstar,
        rv_semiamp_orbit,
        A_noplanet,
        do_pca=do_pca,
        n_princ_comp=n_princ_comp,
        star=star,
        observation="transmission",
    )[0]
    return prior_val + ll


# @numba.njit
def prior(x, best_kp):
    """
    Prior on the parameters. Only uniform!

    Inputs
    ------
        :x: (array) array of parameters

    Outputs
    -------
        :prior_val: (float) log prior value.
    """
    Kp, Vsys, log_scale = x
    # do I sample in log_scale?
    if (
        best_kp - 50.0 < Kp < best_kp + 50.0
        and -50.0 < Vsys < 50.0
        and -1 < log_scale < 1
    ):
        return 0
    return -np.inf


def sample(
    nchains,
    nsample,
    A_noplanet,
    Fp_conv,
    wl_cube_model,
    n_order,
    n_exposure,
    n_pixel,
    star,
    n_princ_comp,
    flux_cube,
    wl_model,
    Fstar_conv,
    Rp_solar,
    Rstar,
    phases,
    do_pca=True,
    best_kp=192.06,
    best_vsys=0.0,
    best_log_scale=0.0,
):
    """
    Samples the likelihood. right now, it needs an instantiated best-fit value.

    Inputs
    ------
        :nchains: (int) number of chains
        :nsample: (int) number of samples
        :A_noplanet: (array) array of the no planet spectrum
        :fTemp: (array) array of the stellar spectrum
        :do_pca: (bool) whether to do PCA
        :best_kp: (float) best-fit planet velocity
        :best_vsys: (float) best-fit system velocity
        :best_log_scale: (float) best-fit log scale

    Outputs
    -------
        :sampler: (emcee.EnsembleSampler) the sampler object.
    """
    # todo: make the likelhiood function based on the sampling parameters.

    pos = np.array([best_kp, best_vsys, best_log_scale]) + 1e-2 * np.random.randn(
        nchains, 3
    )

    # Our 'pool' is just an object with a 'map' method which points to mpi_map
    with MPIPool() as pool:
        if not pool.is_master():
            pool.wait()
            sys.exit(0)
        np.random.seed(42)

        nwalkers, ndim = pos.shape

        sampler = emcee.EnsembleSampler(
            nwalkers,
            ndim,
            log_prob,
            args=(
                best_kp,
                wl_cube_model,
                Fp_conv,
                n_order,
                n_exposure,
                n_pixel,
                A_noplanet,
                star,
                n_princ_comp,
                flux_cube,
                wl_model,
                Fstar_conv,
                Rp_solar,
                Rstar,
                phases,
                do_pca,
            ),
            pool=pool,
        )
        sampler.run_mcmc(pos, nsample, progress=True)

    return sampler



"""
tests the run_simulation module.
"""

import os
import sys
import pytest
import numpy as np
import pandas as pd

from scope.run_simulation import *
from scope.utils import *
from scope.tests.conftest import test_inputs, test_baseline_outouts

# Define the path to the test data
test_data_path = os.path.join(os.path.dirname(__file__), "../data")


# make a pytest fixture to store the test data


@pytest.fixture
def test_noblaze_outputs(test_inputs):
    Fp_conv, Fstar_conv, wl_cube_model, wl_model = test_inputs

    scale = 1

    n_exposure = 10
    n_order, n_pixel = (44, 1848)  # todo: fix.
    phases = np.linspace(-0.01, 0.01, n_exposure)
    Rp_solar = 0.1
    Rstar = 1.0

    # Test the function
    A_noplanet, flux_cube, flux_cube_nopca, just_tellurics = make_data(
        scale,
        wl_cube_model,
        wl_model,
        Fp_conv,
        Fstar_conv,
        n_order,
        n_exposure,
        n_pixel,
        phases,
        Rp_solar,
        Rstar,
        do_pca=True,
        blaze=False,
        tellurics=True,
        n_princ_comp=4,
        v_sys=0,
        Kp=192.06,
        star=True,
        SNR=0,
        rv_semiamp_orbit=0.3229,
        observation="emission",
        tell_type="data-driven",
        time_dep_tell=False,
        wav_error=False,
        order_dep_throughput=True,
        a=1,  # AU
        u1=0.3,
        u2=0.3,
        LD=True,
        b=0.0,  # impact parameter
        divide_out_of_transit=False,
        out_of_transit_dur=0.1,
    )
    return (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    )


@pytest.fixture
def test_notell_outputs(test_inputs):
    Fp_conv, Fstar_conv, wl_cube_model, wl_model = test_inputs

    scale = 1

    n_exposure = 10
    n_order, n_pixel = (44, 1848)  # todo: fix.
    phases = np.linspace(-0.01, 0.01, n_exposure)
    Rp_solar = 0.1
    Rstar = 1.0

    # Test the function
    A_noplanet, flux_cube, flux_cube_nopca, just_tellurics = make_data(
        scale,
        wl_cube_model,
        wl_model,
        Fp_conv,
        Fstar_conv,
        n_order,
        n_exposure,
        n_pixel,
        phases,
        Rp_solar,
        Rstar,
        do_pca=True,
        blaze=True,
        tellurics=False,
        n_princ_comp=4,
        v_sys=0,
        Kp=192.06,
        star=True,
        SNR=0,
        rv_semiamp_orbit=0.3229,
        observation="emission",
        tell_type="data-driven",
        time_dep_tell=False,
        wav_error=False,
        order_dep_throughput=True,
        a=1,  # AU
        u1=0.3,
        u2=0.3,
        LD=True,
        b=0.0,  # impact parameter
        divide_out_of_transit=False,
        out_of_transit_dur=0.1,
    )
    return (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    )


@pytest.fixture
def test_noisy_outputs(test_inputs):
    Fp_conv, Fstar_conv, wl_cube_model, wl_model = test_inputs

    scale = 1

    n_exposure = 10
    n_order, n_pixel = (44, 1848)  # todo: fix.
    phases = np.linspace(-0.01, 0.01, n_exposure)
    Rp_solar = 0.1
    Rstar = 1.0

    # Test the function
    A_noplanet, flux_cube, flux_cube_nopca, just_tellurics = make_data(
        scale,
        wl_cube_model,
        wl_model,
        Fp_conv,
        Fstar_conv,
        n_order,
        n_exposure,
        n_pixel,
        phases,
        Rp_solar,
        Rstar,
        do_pca=True,
        blaze=True,
        tellurics=True,
        n_princ_comp=4,
        v_sys=0,
        Kp=192.06,
        star=True,
        SNR=250,
        rv_semiamp_orbit=0.3229,
        observation="emission",
        tell_type="data-driven",
        time_dep_tell=False,
        wav_error=False,
        order_dep_throughput=True,
        a=1,  # AU
        u1=0.3,
        u2=0.3,
        LD=True,
        b=0.0,  # impact parameter
        divide_out_of_transit=False,
        out_of_transit_dur=0.1,
    )
    return (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    )


@pytest.fixture
def test_baseline_outputs_take2(test_inputs):
    Fp_conv, Fstar_conv, wl_cube_model, wl_model = test_inputs

    scale = 1

    n_exposure = 10
    n_order, n_pixel = (44, 1848)  # todo: fix.
    phases = np.linspace(-0.01, 0.01, n_exposure)
    Rp_solar = 0.1
    Rstar = 1.0

    # Test the function
    A_noplanet, flux_cube, flux_cube_nopca, just_tellurics = make_data(
        scale,
        wl_cube_model,
        wl_model,
        Fp_conv,
        Fstar_conv,
        n_order,
        n_exposure,
        n_pixel,
        phases,
        Rp_solar,
        Rstar,
        do_pca=True,
        blaze=True,
        tellurics=True,
        n_princ_comp=4,
        v_sys=0,
        Kp=192.06,
        star=True,
        SNR=0,
        rv_semiamp_orbit=0.3229,
        observation="transmission",
        tell_type="data-driven",
        time_dep_tell=False,
        wav_error=False,
        order_dep_throughput=True,
        a=1,  # AU
        u1=0.3,
        u2=0.3,
        LD=True,
        b=0.0,  # impact parameter
        divide_out_of_transit=False,
        out_of_transit_dur=0.1,
    )
    return (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    )


def test_make_data_shape(test_baseline_outouts):
    """
    Tests the make_data function.
    """

    (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    ) = test_baseline_outouts

    # assert that the shape is as expected
    assert flux_cube.shape == (n_order, n_exposure, n_pixel)


def test_make_data_not_all_zeros(test_baseline_outouts):
    """
    Tests the make_data function.
    """

    (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    ) = test_baseline_outouts

    # assert that the shape is as expected
    assert not np.all(flux_cube == 0.0)


# def test_tellurics_cube_normalized_correctly(test_baseline_outouts):
#     """
#     Tests the make_data function.
#     """
#
#     A_noplanet, flux_cube, flux_cube_nopca, just_tellurics, n_exposure, n_order, n_pixel = test_baseline_outouts
#
#     # assert that the shape is as expected
#     assert np.max(just_tellurics) == 1.0


def test_flux_values_reasonable(test_baseline_outouts):
    """
    Tests the make_data function.
    """

    (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    ) = test_baseline_outouts

    # assert that the shape is as expected
    assert np.median(flux_cube_nopca) > 0.1


def test_tellurics_regular_values(test_baseline_outouts):
    """
    Tests the make_data function.
    """

    (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    ) = test_baseline_outouts

    # assert that the shape is as expected
    assert np.median(just_tellurics) > 0.1


def test_tellurics_actually_are_penalized(test_baseline_outouts):
    """
    Tests the make_data function.
    """

    (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    ) = test_baseline_outouts

    # assert that the shape is as expected
    assert np.median(just_tellurics) < 1.0


def test_all_finite_made_data(test_baseline_outouts):
    """
    Tests the make_data function.
    """

    (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    ) = test_baseline_outouts

    # assert that the shape is as expected
    assert np.all(np.isfinite(flux_cube))


def test_nopca_fluxlevels_blaze(test_noblaze_outputs, test_baseline_outouts):
    """
    Tests the make_data function.
    """

    (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    ) = test_noblaze_outputs
    (
        A_noplanet_b,
        flux_cube_b,
        flux_cube_nopca_b,
        just_tellurics_b,
        n_exposure_b,
        n_order_b,
        n_pixel_b,
    ) = test_baseline_outouts

    # the flux levels must be higher when there's no blaze.
    assert np.median(flux_cube_nopca) > np.median(flux_cube_nopca_b)


def test_nopca_fluxlevels_tell(test_notell_outputs, test_baseline_outouts):
    """
    Tests the make_data function.
    """

    (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    ) = test_notell_outputs
    (
        A_noplanet_b,
        flux_cube_b,
        flux_cube_nopca_b,
        just_tellurics_b,
        n_exposure_b,
        n_order_b,
        n_pixel_b,
    ) = test_baseline_outouts

    # the flux levels must be higher when there's no tellurics.
    assert np.median(flux_cube_nopca) > np.median(flux_cube_nopca_b)


def test_nopca_fluxlevels_tell(test_notell_outputs, test_baseline_outouts):
    """
    Tests the make_data function.
    """

    (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    ) = test_notell_outputs
    (
        A_noplanet_b,
        flux_cube_b,
        flux_cube_nopca_b,
        just_tellurics_b,
        n_exposure_b,
        n_order_b,
        n_pixel_b,
    ) = test_baseline_outouts

    # the flux levels must be higher when there's no tellurics.
    assert np.median(flux_cube_nopca) > np.median(flux_cube_nopca_b)


def test_nopca_fluxlevels_noisier(test_noisy_outputs, test_baseline_outouts):
    """
    Tests the make_data function.
    """

    (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    ) = test_noisy_outputs
    (
        A_noplanet_b,
        flux_cube_b,
        flux_cube_nopca_b,
        just_tellurics_b,
        n_exposure_b,
        n_order_b,
        n_pixel_b,
    ) = test_baseline_outouts

    # the flux levels must be higher when there's no tellurics.
    assert np.std(flux_cube) > np.std(flux_cube_b)


def test_nopca_fluxlevels_deterministic(
    test_baseline_outputs_take2, test_baseline_outouts
):
    """
    without any noise specified, everything should be deterministic.
    """

    (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    ) = test_baseline_outputs_take2
    (
        A_noplanet_b,
        flux_cube_b,
        flux_cube_nopca_b,
        just_tellurics_b,
        n_exposure_b,
        n_order_b,
        n_pixel_b,
    ) = test_baseline_outouts

    # the flux levels must be higher when there's no tellurics.
    np.testing.assert_array_equal(flux_cube_nopca, flux_cube_nopca_b)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SImMER documentation build configuration file, created by
# sphinx-quickstart on Sat Feb 22 22:19:42 2020.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import codecs
import os
import re
import sys

import mock

import scope

# MOCK_MODULES = [
#     "numpy",
#     "scipy",
#     "sklearn",
#     "matplotlib",
#     "matplotlib.pyplot",
#     "scipy.interpolate",
#     "scipy.special",
#     "math",
#     "__future__",
#     "toolboxutilities",
#     "exoplanet",
#     "sklearn",
#     "astropy.units",
#     "emcee",
#     "jax",
#     "jax.numpy",
#     "pandas",
#     "scipy.constants",
#     "numba",
#     "schwimmbad",
#     "astropy",
#     "astropy.constants",
# ]
# for mod_name in MOCK_MODULES:
#     sys.modules[mod_name] = mock.Mock()

sys.path.insert(0, os.path.abspath("../.."))
sys.path.insert(0, os.path.abspath(".."))
sys.path.insert(0, os.path.abspath("../src"))
sys.path.insert(0, os.path.abspath("../src/scope"))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

META_PATH = os.path.join("..", "src", "scope", "__init__.py")
HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), "rb", "utf-8") as f:
        return f.read()


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))


META_FILE = read(META_PATH)

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "nbsphinx",
    "sphinx_book_theme",
]


nbsphinx_prolog = """
This notebook is available at
https://github.com/arjunsavel/scope/tree/main/docs/{{ env.doc2path(env.docname, base=None) }}
"""
# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

napoleon_google_docstring = False
napoleon_use_param = False
napoleon_use_ivar = True

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "scope"
copyright = "2024, Arjun B. Savel, Megan Bedell, Eliza M.-R. Kempton, Peter Smith, Jacob L. Bean, Lily L. Zhao, Jorge A. Sanchez, and Michael R. Line"
author = "Arjun B. Savel, Megan Bedell, Eliza M.-R. Kempton, Peter Smith, Jacob L. Bean, Lily L. Zhao, Jorge A. Sanchez, and Michael R. Line"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = scope.__version__
# The full version, including alpha/beta/rc tags.
# release = find_meta("release")

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_
# _path and html_extra_path
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "../src/scope/tests",
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"
# html_favicon = "_static/favicon.png"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "scopedoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "scope.tex",
        "scope Documentation",
        "Arjun B. Savel, Megan Bedell, Eliza M.-R. Kempton, Peter Smith, Jacob L. Bean, Lily L. Zhao, Jorge A. Sanchez, and Michael R. Line",
        "manual",
    )
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "scope", "scope Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "scope",
        "scope Documentation",
        author,
        "scope",
        "Miscellaneous",
    )
]



"""
file to calculate the intensity map and broadening said map. and the velocity profile.
"""


import numpy as np
from numba import njit

from scope.constants import *


def get_rot_ker(vsini, wStar):
    """
    Broadens by a single velocity?
    """
    (nx,) = wStar.shape
    dRV = np.mean(2.0 * (wStar[1:] - wStar[0:-1]) / (wStar[1:] + wStar[0:-1])) * 2.998e5
    nker = 401
    hnker = (nker - 1) // 2
    rker = np.zeros(nker)
    for ii in range(nker):
        ik = ii - hnker
        x = ik * dRV / vsini
        if np.abs(x) < 1.0:
            y = np.sqrt(1 - x**2)
            rker[ii] = y
    rker /= rker.sum()

    return rker


# @njit
def get_theta(lat, lon):
    """
    Converts latitude and longitude to Theta, the angular distance across the disk.

    Inputs
    ------
        :lat: latitude in radians
        :lon: longitude in radians

    Outputs
    -------
        :res1: the angular distance across the disk.

    """
    if np.any(np.abs(lat) > np.pi) or np.any(np.abs(lon) > np.pi):
        raise ValueError("lat or lon is too large. are you inputting in radians?")
    else:
        return np.arccos(np.sqrt(1 - np.sin(lon) ** 2 - np.sin(lat) ** 2))


def fix_nan_result(res):
    """
    Fixes the result of a calculation that has NaNs in it.

    Parameters
    ----------
    res : float
        The result of a calculation.

    Returns
    -------
    float
        The result of the calculation with NaNs fixed.
    """
    if isinstance(res, np.ndarray) or isinstance(res, list):
        res[np.isnan(res)] = 0.0
    else:
        if np.isnan(res):
            res = 0.0
    return res


# @njit
def I_darken(lon, lat, epsilon):
    """
    takes lon and lat in. spits out the value on the limb-darkened disk.

    Inputs
    ------
        :lon: longitude in radians
        :lat: latitude in radians
        :epsilon: limb-darkening coefficient. should be between 0 and 1.

    Outputs
    -------
        :res1: the intensity at that point on the disk.
    """
    if 0.0 <= epsilon <= 1.0:
        theta = get_theta(lat, lon)
        res1 = (1 - epsilon) + (epsilon * np.cos(theta))

        res1 = fix_nan_result(res1)

        return res1
    else:
        raise ValueError("epsilon should only be from 0 to 1")


# @njit
def I_darken_disk(x, y, epsilon):
    """
    takes x and y in. spits out the value on the limb-darkened disk.

    Inputs
    ------
        :x: x position on the disk
        :y: y position on the disk
        :epsilon: limb-darkening coefficient. should be between 0 and 1.

    Outputs
    -------
        :res1: the intensity at that point on the disk.
    """
    lon = np.arcsin(x)
    lat = np.arcsin(y)

    return I_darken(lon, lat, epsilon)


def I_darken_disk_integrated(x, y, epsilon):
    """
    takes x and y in. spits out the value on the limb-darkened disk.

    Inputs
    ------
        :x: x position on the disk
        :y: y position on the disk
        :epsilon: limb-darkening coefficient. should be between 0 and 1.

    Outputs
    -------
        :res1: the intensity at that point on the disk.
    """
    lon = np.arcsin(x)
    lat = np.arcsin(y)

    dTheta = dx / np.cos(lon)
    dPhi = dy / np.cos(lat)

    return np.sum(
        dTheta * dPhi * np.cos(lon) * np.cos(lat) * I_darken(lon, lat, epsilon)
    )


@njit
def gaussian_term(lon, lat, offset, sigma, amp):
    """
    the gaussian term.

    Inputs
    ------
        :lat: latitude in radians
        :lon: longitude in radians
        :offset: the offset of the gaussian
        :sigma: the sigma of the gaussian
        :amp:  the amplitude of the gaussian

    Outputs
    -------
        :res1: the gaussian term.

    """
    return gaussian_term_1d(lat, 0.0, sigma, amp) * np.exp(
        -((lon - offset) ** 2) / (2 * sigma**2)
    )


@njit
def gaussian_term_1d(lat, offset, sigma, amp):
    """
    the gaussian term. this time there's no longitude dependence!

    Inputs
    ------
        :lat: latitude in radians
        :lon: longitude in radians
        :offset: the offset of the gaussian
        :sigma: the sigma of the gaussian
        :amp: the amplitude of the gaussian

    Outputs
    -------
        :res1: the gaussian term.
    """
    return (
        amp
        * (1 / (sigma**2 * 2 * np.pi))
        * np.exp(-((lat - offset) ** 2) / (2 * sigma**2))
    )


#### below: the actual integrations

vl = 1


@njit
def numerator_integral_no_sigma(vz, epsilon, n_samples=100):
    """
    calculates the numerator of the line profile. this is the version without the exponent.

    Parameters
    ----------
    vz : float
        The velocity of the star.
    epsilon : float
        The limb-darkening coefficient.
    n_samples : int
        The number of samples to use in the integration.

    Returns
    -------
    total_res : float
        The numerator of the line profile.
    """
    # set bounds
    lower_bound = 0
    upper_bound = np.arcsin(np.sqrt(1 - np.square(vz / vl)))
    # need to fix these bounds
    lat_arr = np.linspace(lower_bound, upper_bound, n_samples)
    d_lat = np.diff(lat_arr)[0]

    total_res = 0.0
    for lat in lat_arr:
        res = (
            (
                (1 - epsilon)
                + epsilon * np.sqrt(1 - np.square(vz / vl) - np.square(np.sin(lat)))
            )
            * np.cos(lat)
            * d_lat
        )
        if not np.isnan(res):
            total_res += res
    return total_res


@njit
def numerator_no_sigma(vz, epsilon, n_samples=100):
    """
    calculates the numerator of the line profile. this is the version without the exponent.

    Parameters
    ----------
    vz : float
        The velocity of the star.
    epsilon : float
        The limb-darkening coefficient.
    n_samples : int
        The number of samples to use in the integration.

    Returns
    -------
    res1 : float
        The numerator of the line profile.
    """
    res1 = 2
    return res1 * numerator_integral_no_sigma(vz, epsilon, n_samples=n_samples)


# @njit
def line_profile_no_exponent(epsilon, dRV, n_samples=int(1e4), model="igrins", vl=5):
    """
    calculates the line profile for a given epsilon and dRV. this is the version without the exponent.

    Parameters
    ----------
    epsilon : float
        The limb-darkening coefficient.
    dRV : float
        The width of the line profile.
    n_samples : int
        The number of samples to use in the integration.
    model : str
    vl : float
        The equatorial rotational velocity of the star.

    Returns
    -------
    big_resoluts : np.ndarray
        The line profile.

    """
    if model == "igrins":
        n_vz = 2 * vl / dRV
    else:
        n_vz = 500
    vzs = np.linspace(
        -1, 1, int(n_vz)
    )  # need to set to 0 if off disk. this is actually vz / vl. can also step off.
    big_resoluts = np.zeros(vzs.shape)

    for i, vz in enumerate(vzs):
        re_num = numerator_no_sigma(vz, epsilon, n_samples=n_samples)
        #         vz, epsilon, mu, sigma=.7, n_samples=100

        # todo: check sensitivity to lat / lon gridding
        re_denom = I_darken_disk_integrated(X, Y, epsilon)
        # pdb.set_trace()
        big_resoluts[i] = re_num / re_denom

    # normalize
    big_resoluts /= np.sum(big_resoluts)
    return big_resoluts


def convert_vz_to_lon(vz, R, sigma_jet, amp_jet, mu_jet):
    """
    Converts vz to a longitude.

    Parameters
    ----------
    vz : float
        The velocity along the line of sight.
    R : float
        The radius of the star.
    sigma_jet : float
        The width of the jet.
    amp_jet : float
        The amplitude of the jet.
    mu_jet : float
        The center of the jet.


    """
    gauss_term = gaussian_term_1d(vz, sigma_jet, amp_jet, mu_jet)
    return np.arcsin(vz / (R * gauss_term))


def convert_lon_to_vz(lon, R, sigma_jet, amp_jet, mu_jet):
    """
    Converts longitude to a vz

    Parameters
    ----------
    lon : float
        The longitude.
    R : float
        The radius of the planet.
    sigma_jet : float
        The width of the jet.
    amp_jet : float
        The amplitude of the jet.
    mu_jet : float
        The center of the jet.


    """
    vz = (
        R * np.sin(lon) * gaussian_term_1d(lon, sigma_jet, amp_jet, mu_jet)
    )  # todo: check another factor of R?
    return vz


# @njit
def broaden_spectrum(
    wav, spectrum_flux, epsilon, n_samples=int(1e4), vl=5, model="igrins"
):
    """
    Broadens a spectrum by a line profile.

    Parameters
    ----------
    wl_model : array-like
        The wavelength model.
    spectrum_flux : array-like
        The flux of the spectrum.
    epsilon : float
        The spot contrast.
    n_samples : int
        The number of samples to use in the integration.
    vl : float
        The rotational velocity of the object.

    Returns
    -------
    array-like
        The broadened spectrum.
    """
    # pdb.set_trace()
    dRV = (
        np.mean(
            np.array(
                [np.mean(np.diff(wav) / wav[1:]), np.mean(np.diff(wav) / wav[:-1])]
            )
        )
        * const_c
        / 1e3
    )
    profile = line_profile_no_exponent(
        epsilon, dRV, n_samples=n_samples, vl=vl, model=model
    )
    return np.convolve(spectrum_flux, profile, mode="same")


x = np.linspace(-1, 1, 80)
y = np.linspace(-1, 1, 80)

X, Y = np.meshgrid(x, y)

dx = np.diff(x)[0]
dy = np.diff(y)[0]

if __name__ == "__main__":
    x = np.linspace(-1, 1, 80)
    y = np.linspace(-1, 1, 80)

    vl = 5

    X, Y = np.meshgrid(x, y)

    dx = np.diff(x)[0]
    dy = np.diff(y)[0]

    lon = np.linspace(-90, 90, 80)
    lat = np.linspace(-90, 90, 80)

    lon = np.radians(lon)
    lat = np.radians(lat)

    LON, LAT = np.meshgrid(lon, lat)

    x = np.linspace(-1, 1, 80)
    y = np.linspace(-1, 1, 80)

    X, Y = np.meshgrid(x, y)

    res = I_darken(LON, LAT, 0.1)
    res_disk = I_darken_disk(X, Y, 0.9)



# -*- coding: utf-8 -*-
try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:  # for Python<3.8
    from importlib_metadata import PackageNotFoundError, version

try:
    __version__ = version("scope")
except PackageNotFoundError:
    # package is not installed
    from ._version import __version__  # type: ignore



"""
This file tests the cross-correlation function.
"""

import numpy as np
import pytest

from scope.ccf import calc_ccf


def test_autocorr_peaks_zero():
    """
    Tests that the autocorrelation peaks at zero.
    """
    n_pixel = 101
    x = np.arange(n_pixel)
    y = np.sin(x)
    model_flux = np.array([y, y])
    # normalize arrays correctly. subtract mean and divide standard deviation
    model_flux = (model_flux - np.mean(model_flux, axis=1)[:, None]) / np.std(
        model_flux, axis=1
    )[:, None]

    # now shift the data in a loop to find the maximum CCF

    model_flux_slice = model_flux.copy()

    CCF_arr = np.ones(n_pixel)
    for i in range(0, n_pixel):
        model_flux_slice = np.roll(model_flux, i, axis=1)
        logl, CCF = calc_ccf(model_flux, model_flux_slice, n_pixel)
        CCF_arr[i] = CCF

    assert np.argmax(CCF_arr) == 0



"""
Read in and apply tellurics to simulated data.
"""

import os
from glob import glob

import numpy as np
import pandas as pd
from scipy import interpolate

abs_path = os.path.dirname(__file__)


def read_atran(path):
    """
    reads a single ATRAN file.

    Inputs
    ------
    path: string
        Path to the file.

    Returns
    -------
    atran: dataframe
        Atran file.
    """
    atran = pd.read_csv(
        path,
        index_col=0,
        names=["wav", "depth"],
        sep=r"\s+",
    ).drop_duplicates(subset=["wav"])

    return atran


def read_atran_pair(path1, path2):
    """
    reads and concatenates an ATRAN file pair.
    """
    atran_40_first_half = read_atran(path1)

    atran_40_second_half = read_atran(path2)

    atran_40 = pd.concat([atran_40_first_half, atran_40_second_half]).drop_duplicates(
        subset=["wav"]
    )
    return atran_40


def find_closest_atran_angle(zenith_angle):
    """
    Parses through the available ATRAN files and finds the one with the zenith angle closest to the desired.
    """
    files = glob("data/atran*obslat*")
    zenith_paths = np.array([eval(file.split("_")[1]) for file in files])
    closest_ind = np.argmin(abs(zenith_paths - zenith_angle))
    closest_zenith = zenith_paths[closest_ind]
    return closest_zenith


def add_tellurics_atran(
    wl_cube_model, flux_cube_model, n_order, n_exp, vary_airmass=False
):
    if vary_airmass:
        # assume now that I'm just using the same airmasses as before.
        zenith_angles = np.loadtxt("zenith_angles_w77ab.txt")

        # these are the beginning and ends of the two different ATRAN chunks.
        w_start_1 = 1.1
        w_end_1 = 2
        w_start_2 = 2
        w_end_2 = 3

        tell_splines = []  # will want one of these for each exposure.
        for zenith_angle in zenith_angles:
            zenith_path = find_closest_atran_angle(zenith_angle)
            path1 = f"data/atran_{zenith_path}_zenith_39_obslat_{w_start_1}_{w_end_1}_wave.dat"
            path2 = f"data/atran_{zenith_path}_zenith_39_obslat_{w_start_2}_{w_end_2}_wave.dat"
            atran_40 = read_atran_pair(path1, path2)
            tell_spline = interpolate.splrep(atran_40.wav, atran_40.depth, s=0.0)
            tell_splines += [tell_spline]

        for order in tqdm(range(n_order)):
            wl_grid = wl_cube_model[order]
            for exp in range(n_exp):
                tell_spline = tell_splines[exp]
                tell_flux = interpolate.splev(wl_grid, tell_spline, der=0)

                flux_cube_model[order, exp] *= tell_flux
    else:
        path1 = "atran_45k_1.3_2_microns_40_deg.txt"
        path2 = "atran_45k_2_2.7_microns_40_deg.txt"
        atran_40 = read_atran_pair(path1, path2)

        # todo: refactor. there are a lot of similarities!
        tell_spline = interpolate.splrep(atran_40.wav, atran_40.depth, s=0.0)
        for order in tqdm(range(n_order)):
            wl_grid = wl_cube_model[order]
            for exp in range(n_exp):
                tell_flux = interpolate.splev(wl_grid, tell_spline, der=0)
                flux_cube_model[order, exp] *= tell_flux

    return flux_cube_model


def calc_weighted_vector_insim(
    vals, eigenvector, relationships, provided_relationships=False
):
    """
    Calculates the weighted vector for a given set of eigenvalues and eigenvectors.

    Inputs
    ------
        :vals: (array) eigenvalues
        :eigenvector: (array) eigenvectors
        :relationships: (list of arrays) relationships between eigenvectors and the flux vector.
        :provided_relationships: (bool) if True, then relationships are provided. If False, then relationships are calculated.

    Outputs
    -------
        :fm: (array) weighted vector
        :relationships: (list of arrays) relationships between eigenvectors and the flux vector.
    """

    fm = np.zeros(1698)  # pad with 1s on all sides.
    if not provided_relationships:
        relationships = []
    for i, vector_ind in enumerate(range(4)):
        relationship = relationships[i]

        vector = eigenvector[:, vector_ind]

        eigenweight = np.dot(relationship, vals)

        weighted_vector = eigenweight * vector

        fm += weighted_vector
        if not provided_relationships:
            relationships += [relationship]
    fm[fm < 0.0] = 0.0
    return fm, relationships


def eigenweight_func(airmass, date):
    """
    Does the weighting. turns date and airmass into their correct array!

    Inputs
    ------
        :airmass: (float) airmass
        :date: (float) date

    Outputs
    -------
        :array: (array) array of values to be multiplied by the eigenvectors.
    """
    if airmass < 1.0 or airmass > 3.0:
        raise ValueError(
            f"Airmass must be valued between 1 amd 3. given airmass is {airmass}"
        )
    if date < -200 or date > 400:
        raise ValueError
    return np.array([1, airmass, date, date**2])


def add_tellurics(
    wl_cube_model,
    flux_cube_model,
    n_order,
    n_exp,
    vary_airmass=False,
    tell_type="ATRAN",
    time_dep_tell=False,
):
    """
    Includes tellurics in the model.
    todo: allow thew airmass variation to not be the case for the data-driven tellurics

    Inputs
    ------
        :wl_cube_model: (array) wavelength cube model
        :flux_cube_model: (array) flux cube model
        :n_order: (int) number of orders
        :n_exp: (int) number of exposures
        :vary_airmass: (bool) if True, then the airmass is varied. If False, then the airmass is fixed.
        :tell_type: (str) either 'ATRAN' or 'data-driven'
        :time_dep_tell: (bool) if True, then the tellurics are time dependent. If False, then the tellurics are not time dependent.

    Outputs
    -------
        :flux_cube_model: (array) flux cube model with tellurics included.

    """
    if tell_type == "data-driven":
        eigenvectors = np.load(abs_path + "/data/eigenvectors.npy")
        relationships_arr = np.load(abs_path + "/data/eigenweight_coeffs.npy")
        wave_for_eigenvectors = np.load(abs_path + "/data/wav_for_eigenvectors.npy")

        zenith_angles = np.loadtxt(abs_path + "/data/zenith_angles_w77ab.txt")

        airmasses = 1 / np.cos(np.radians(zenith_angles))  # todo: check units
        airmasses = airmasses[:n_exp]
        if time_dep_tell:
            # dates = np.loadtxt('dates_w77ab_scaled.txt')
            dates = np.linspace(0, 350, len(airmasses))
        else:
            dates = np.ones_like(airmasses)

        # iterate through the each exposure.
        for i, airmass in enumerate(airmasses):
            date = dates[i]
            for order in range(relationships_arr.shape[0]):
                vals = eigenweight_func(airmass, date)
                fm, _ = calc_weighted_vector_insim(
                    vals,
                    eigenvectors[order],
                    relationships_arr[order],
                    provided_relationships=True,
                )
                fm = np.concatenate([np.ones(150), fm])

                flux_cube_model[order][
                    i
                ] *= fm  # the orders don't quite line up. or maybe they do : )

    elif tell_type == "ATRAN":
        flux_cube_model = add_tellurics_atran(
            wl_cube_model, flux_cube_model, n_order, n_exp, vary_airmass=vary_airmass
        )
    return flux_cube_model



"""
Add noise to simulated data.
"""
import numpy as np

from scope.utils import *

test_data_path = os.path.join(os.path.dirname(__file__), "data")


def add_constant_noise(flux_cube_model, wl_grid, SNR):
    """
    Adds constant noise to the flux cube model.

    Inputs
    ------
        :flux_cube_model: (array) flux cube model
        :wl_grid: (array) wavelength grid
        :SNR: (float) signal-to-noise ratio

    Outputs
    -------
        :noisy_flux: (array) flux cube model with constant noise added.
    """
    n_photons = SNR**2
    # SNR**2 is the total number of photons.
    flux_cube = (
        flux_cube_model.copy() * n_photons
    )  # now scaled by number of photons. it's the maximum

    noise_matrix = np.random.normal(loc=0, scale=1, size=flux_cube_model.shape)
    noise_matrix_scaled = noise_matrix * np.sqrt(flux_cube)
    noisy_flux = flux_cube + noise_matrix_scaled
    return noisy_flux


def add_quadratic_noise(flux_cube_model, wl_grid, SNR, IGRINS=False, **kwargs):
    """
    Currently assumes that there are two bands: A and B.

    Inputs
    ------
        :flux_cube_model: (array) flux cube model
        :wl_grid: (array) wavelength grid
        :SNR: (float) signal-to-noise ratio
        :IGRINS: (bool) if True, then the data is IGRINS data. If False, then the data is NOT IGRINS data.

    Outputs
    -------
        :noisy_flux: (array) flux cube model with quadratic noise added.
    """
    if IGRINS:
        A_a, A_b, A_c, B_a, B_b, B_c = np.loadtxt(
            os.path.join(test_data_path, "igrins_median_snr.txt")
        )
        # this is scaled to a mean SNR of 250.
        noisy_flux = np.ones_like(flux_cube_model) * 0
        wl_cutoff = 1.9
        medians = np.median(wl_grid, axis=1)

        A_wl = medians[medians < wl_cutoff]
        B_wl = medians[medians > wl_cutoff]

        A_SNRs = A_a + A_b * A_wl + A_c * A_wl**2
        B_SNRs = B_a + B_b * B_wl + B_c * B_wl**2

        # need to scale these to the required SNR.
        A_SNRs *= SNR / 250
        B_SNRs *= SNR / 250
        noisy_flux = np.ones_like(flux_cube_model) * 0

        # iterate through each exposure
        for exposure in range(flux_cube_model.shape[1]):
            for order in range(flux_cube_model.shape[0]):
                if medians[order] < wl_cutoff:  # first band
                    order_snr = A_SNRs[order - len(B_SNRs)]
                else:  # second band
                    order_snr = B_SNRs[order]
                flux_level = (
                    (order_snr**2)
                    * flux_cube_model[order][exposure]
                    / np.nanmax(flux_cube_model[order][exposure])
                )
                noisy_flux[order][exposure] = np.random.poisson(flux_level)

                # a few quick checks to make sure that nothing has gone wrong with adding noise
                noisy_flux[order][exposure][noisy_flux[order][exposure] < 0.0] = 0.0
                noisy_flux[order][exposure][
                    ~np.isfinite(noisy_flux[order][exposure])
                ] = 0.0

    else:
        raise NotImplementedError("Only IGRINS data is currently supported.")

    return noisy_flux


def add_igrins_noise(flux_cube_model, wl_grid, SNR):
    """
    Adds IGRINS noise to the flux cube model.

    Inputs
    ------
        :flux_cube_model: (array) flux cube model
        :wl_grid: (array) wavelength grid
        :SNR: (float) signal-to-noise ratio

    Outputs
    -------
        :noisy_flux: (array) flux cube model with IGRINS noise added.
    """
    return add_quadratic_noise(flux_cube_model, wl_grid, SNR, IGRINS=True)


def add_custom_noise(SNR):
    """
    Adds custom noise to the flux cube model.
    """
    raise NotImplementedError("Custom noise is not yet implemented.")


def add_noise_cube(flux_cube_model, wl_grid, SNR, noise_model="constant", **kwargs):
    """
    Per the equation in Brogi + Line 19.
    Assumes that the flux cube is scaled 0 to 1.

    I guess note that when I say "SNR", I'm ignoring for now the wavelength-dependent
    flux that you get because of a blaze function. It's the SNR at the peak of the blaze function.

    Inputs
    ------
        :flux_cube_model: (array) flux cube model
        :wl_grid: (array) wavelength grid
        :SNR: (float) signal-to-noise ratio
        :noise_model: (str) noise model to use. Can be constant, IGRINS, custom_quadratic, or custom.

    Outputs
    -------
        :noisy_flux: (array) flux cube model with noise added.
    """

    noise_models = {
        "constant": add_constant_noise,
        "IGRINS": add_igrins_noise,
        "custom_quadratic": add_quadratic_noise,
        "custom": add_custom_noise,
    }
    if noise_model not in noise_models.keys():
        raise ValueError(
            "Noise model can only be constant, IGRINS, custom_quadratic, or custom."
        )

    noise_func = noise_models[noise_model]

    noisy_flux = noise_func(flux_cube_model, wl_grid, SNR, **kwargs)

    # need to make sure that it's still normed 0 to 1!
    noisy_flux = detrend_cube(noisy_flux, noisy_flux.shape[0], noisy_flux.shape[1])

    return noisy_flux



import os
import tempfile

import numpy as np
import pytest

from scope.input_output import (  # Replace 'your_module' with the actual module name
    parse_input_file,
    write_input_file,
    ScopeConfigError,
)


@pytest.fixture
def sample_files():
    # Create a temporary directory for test files
    with tempfile.TemporaryDirectory() as test_dir:
        # Create a sample input file
        input_file_content = """:········································
:                                       :
:    ▄▄▄▄▄   ▄█▄    ████▄ █ ▄▄  ▄███▄   :
:   █     ▀▄ █▀ ▀▄  █   █ █   █ █▀   ▀  :
: ▄  ▀▀▀▀▄   █   ▀  █   █ █▀▀▀  ██▄▄    :
:  ▀▄▄▄▄▀    █▄  ▄▀ ▀████ █     █▄   ▄▀ :
:            ▀███▀         █    ▀███▀   :
:                           ▀           :
:                                       :
:········································
Created:                2024-08-15
Author:                 Arjun Savel!
Planet name:            GJ 1214b

# Astrophysical Parameters
Rp                     1.5
Rp_solar               DATABASE
Rstar                  NULL
kp                     150.0
v_rot                  5.0
v_sys                  0.0

# Instrument Parameters
blaze                  True          # whether to include a blaze function or not.
wav_error              False         # whether to include wavelength solution errors or not.
order_dep_throughput   True          # whether to include order-dependent throughput variations.

# Observation Parameters
observation            emission
phase_start                 0.3
phase_end                 0.5
blaze                  True
star                   False
tell_type              data-driven   # type of telluric simulation. supported modes are ``ATRAN`` and ``data-driven``.
time_dep_tell          False         # whether the tellurics are time-dependent or not.
"""
        input_file_path = os.path.join(test_dir, "test_input.txt")
        with open(input_file_path, "w") as f:
            f.write(input_file_content)

        # Create a sample database file
        db_content = """
pl_name,planet_radius_solar
GJ 1214b,0.15
"""
        db_file_path = os.path.join(test_dir, "test_db.csv")
        with open(db_file_path, "w") as f:
            f.write(db_content)

        yield input_file_path, db_file_path


def test_parse_input_file(sample_files):
    input_file_path, db_file_path = sample_files
    data = parse_input_file(input_file_path, db_file_path)

    assert data["planet_name"] == "GJ 1214b"
    assert data["Rp"] == 1.5
    assert np.isnan(data["Rstar"])
    assert data["kp"] == 150.0
    assert data["v_rot"] == 5.0
    assert data["v_sys"] == 0.0
    assert data["observation"] == "emission"
    assert data["phase_start"] == 0.3
    assert data["phase_end"] == 0.5
    assert data["blaze"] == True
    assert data["star"] == False


def test_write_input_file(sample_files, tmp_path):
    input_file_path, db_file_path = sample_files

    # First, parse the input file
    data = parse_input_file(input_file_path, db_file_path)

    # Write the data to a new file
    output_file_path = tmp_path / "output_input.txt"
    write_input_file(data, str(output_file_path))

    # Now parse the output file and compare with original data
    new_data = parse_input_file(str(output_file_path), db_file_path)

    # Compare key elements (you might want to expand this)
    assert new_data["planet_name"] == data["planet_name"]
    assert new_data["Rp"] == data["Rp"]
    assert np.isnan(new_data["Rstar"])
    assert new_data["kp"] == data["kp"]
    assert new_data["observation"] == data["observation"]
    assert new_data["phase_start"] == data["phase_start"]
    assert new_data["blaze"] == data["blaze"]
    assert new_data["star"] == data["star"]


def test_error_on_data_driven_tell(sample_files):
    input_file_path, db_file_path = sample_files
    data = parse_input_file(input_file_path, db_file_path)
    data["tell_type"] = "data-driven"
    data["blaze"] = False

    test_file = "test_input_broken_tell.txt"
    write_input_file(data, test_file)
    # now reading it in should raise the error
    with pytest.raises(ScopeConfigError) as exc:
        parse_input_file(test_file, db_file_path)

    assert "Data-driven tellurics requires blaze set to True." in str(exc.value)



import unittest

from scope.constants import *


class TestC(unittest.TestCase):
    def test_c_large(self):
        self.assertTrue(const_c > 2e8)

    def test_c_small(self):
        self.assertTrue(const_c < 3e8)


class TestRjup(unittest.TestCase):
    def test_rjup_small(self):
        self.assertTrue(rjup_rsun < 0.2)

    def test_rjup_large(self):
        self.assertTrue(rjup_rsun > 0.1)



import os
import sys
import pytest
import numpy as np
import pandas as pd

from scope.run_simulation import *
from scope.utils import *

# Define the path to the test data
test_data_path = os.path.join(os.path.dirname(__file__), "../data")


@pytest.fixture
def test_inputs():
    """
    sets up the spectral things.
    """
    data_cube_path = os.path.join(test_data_path, "data_RAW_20201214_wl_algn_03.pic")
    planet_spectrum_path = os.path.join(
        test_data_path, "best_fit_spectrum_tran_gj1214_steam_nir.pic"
    )
    wl_model, Fp, Fstar = np.load(planet_spectrum_path, allow_pickle=True)
    mike_wave, mike_cube = pickle.load(open(data_cube_path, "rb"), encoding="latin1")
    wl_cube_model = mike_wave.copy().astype(np.float64)

    v_rot = 4.33  # km/s
    rot_ker = get_rot_ker(v_rot, wl_model)
    Fp_conv_rot = np.convolve(Fp, rot_ker, mode="same")
    # instrument profile convolution
    instrument_kernel = get_instrument_kernel()
    Fp_conv = np.convolve(Fp_conv_rot, instrument_kernel, mode="same")
    star_spectrum_path = os.path.join(test_data_path, "PHOENIX_5605_4.33.txt")
    star_wave, star_flux = np.loadtxt(
        star_spectrum_path
    ).T  # Phoenix stellar model packing
    Fstar_conv = get_star_spline(
        star_wave, star_flux, wl_model, instrument_kernel, smooth=False
    )

    return Fp_conv, Fstar_conv, wl_cube_model, wl_model


@pytest.fixture
def test_baseline_outouts(test_inputs):
    Fp_conv, Fstar_conv, wl_cube_model, wl_model = test_inputs

    scale = 1

    n_exposure = 10
    n_order, n_pixel = (44, 1848)  # todo: fix.
    phases = np.linspace(-0.01, 0.01, n_exposure)
    Rp_solar = 0.1
    Rstar = 1.0

    # Test the function
    A_noplanet, flux_cube, flux_cube_nopca, just_tellurics = make_data(
        scale,
        wl_cube_model,
        wl_model,
        Fp_conv,
        Fstar_conv,
        n_order,
        n_exposure,
        n_pixel,
        phases,
        Rp_solar,
        Rstar,
        do_pca=True,
        blaze=True,
        tellurics=True,
        n_princ_comp=4,
        v_sys=0,
        Kp=192.06,
        star=True,
        SNR=0,
        rv_semiamp_orbit=0.3229,
        observation="transmission",
        tell_type="data-driven",
        time_dep_tell=False,
        wav_error=False,
        order_dep_throughput=True,
        a=1,  # AU
        u1=0.3,
        u2=0.3,
        LD=True,
        b=0.0,  # impact parameter
        divide_out_of_transit=False,
        out_of_transit_dur=0.1,
    )
    return (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    )



import pytest
import os
from scipy import signal
from scipy.optimize import curve_fit

from scope.rm_effect import *

test_data_path = os.path.join(os.path.dirname(__file__), "../data")


@pytest.mark.parametrize(
    "values, output",
    [
        ([10, 10], (100, 2)),
        ([10, 1], (10, 2)),
        ([23, 2], (46, 2)),
    ],
)
def test_make_grid_shape(values, output):
    """
    Tests that the grid is the correct shape.
    """
    nr, ntheta = values
    grid = make_grid(nr, ntheta)

    assert grid.shape == output


@pytest.fixture
def test_make_grid_values():
    """
    Tests that the grid is the correct shape.
    """
    nr, ntheta = 10, 10
    grid = make_grid(nr, ntheta)

    return grid


def test_calc_areas_shape(test_make_grid_values):
    grid = test_make_grid_values
    areas = calc_areas(grid)
    assert areas.shape == (100,)


def test_calc_areas_all_nonzerro(test_make_grid_values):
    grid = test_make_grid_values
    areas = calc_areas(grid)
    assert np.all(areas > 0)


def test_calc_planet_locations_shape():
    phases = [1]
    r_star = 1
    a = 2
    inc = np.pi / 2
    lambda_misalign = 0
    res = calc_planet_locations(phases, r_star, inc, lambda_misalign, a)
    assert res.shape == (len(phases), 2)


@pytest.mark.parametrize(
    "values, output",
    [
        ([[0], 1, np.pi / 2, 0, 3], [[0, np.pi / 2]]),  # center of star at phase 0.
        (
            [[0], 1, np.pi / 2, 0, 3e3],
            [[0, np.pi / 2]],
        ),  # center of star at phase 0 even with a huge semimajor axis
        (
            [[0], 1, np.pi / 2, np.pi / 2, 3e3],
            [[0, np.pi]],
        ),  # center of star at phase 0 even with lambda misalignment
        (
            [[0], 1, np.pi / 2, np.pi / 2, 3e3],
            [[0, np.pi]],
        ),  # center of star at phase 0 even with lambda misalignment
        (
            [[0], 1, np.radians(85), np.pi / 2, 3e3],
            [[3e3 * np.cos(np.radians(85)), np.pi]],
        ),  # get a little bit of y.
        (
            [[0], 2, np.radians(85), np.pi / 2, 3e3],
            [[3e3 * np.cos(np.radians(85)) / 2, np.pi]],
        ),  # get a little bit of y, varying rstar.
        (
            [[0.5], 2, np.radians(85), np.pi / 2, 3e3],
            [[3e3 * np.cos(np.radians(85)) / 2, 0]],
        ),  # same answer at phase 0.5
        (
            [[0.5], 2, np.radians(85), 0, 3e3],
            [[3e3 * np.cos(np.radians(85)) / 2, -np.pi / 2]],
        ),  # and it flips if misaligntment woohoo
    ],
)
def test_calc_planet_locations_center_of_transit(values, output):
    # oh the positions are in r and theta lol
    phases, r_star, inc, lambda_misalign, a = values
    res = calc_planet_locations(phases, r_star, inc, lambda_misalign, a)
    assert np.allclose(res, output)


@pytest.fixture
def test_load_phoenix_model():
    star_spectrum_path = os.path.join(test_data_path, "PHOENIX_5605_4.33.txt")
    star_wave, star_flux = np.loadtxt(
        star_spectrum_path
    ).T  # Phoenix stellar model packing
    star_wave *= 1e-6  # convert to meters
    return star_wave, star_flux


def test_doppler_shift_grid_shape(test_make_grid_values, test_load_phoenix_model):
    grid = test_make_grid_values
    star_wave, star_flux = test_load_phoenix_model
    n_wl = len(star_wave)

    res = doppler_shift_grid(grid, star_flux, star_wave, 3)
    assert res.shape == (grid.shape[0], n_wl)


def test_doppler_shift_grid_approach_side_more_blue(
    test_make_grid_values, test_load_phoenix_model
):
    grid = test_make_grid_values
    star_wave, star_flux = test_load_phoenix_model
    n_wl = len(star_wave)

    spectrum_grid = doppler_shift_grid(grid, star_flux, star_wave, 3)

    # from the approach side, the star should be more blue.
    max_r = np.max(grid[:, 0])
    most_right = grid[:, 1][np.argmin(np.abs(grid[:, 1] - 0))]
    most_left = grid[:, 1][np.argmin(np.abs(grid[:, 1] - np.pi))]

    # get the gridpoint that is most right and max r

    most_right_idx = np.argwhere((grid[:, 0] == max_r) & (grid[:, 1] == most_right))[0][
        0
    ]
    most_left_idx = np.argwhere((grid[:, 0] == max_r) & (grid[:, 1] == most_left))[0][0]

    # now get the spectra at these points from the spectrum grid.

    most_right_spec = spectrum_grid[most_right_idx]
    most_left_spec = spectrum_grid[most_left_idx]

    # now, the right one should be redshifted compared to the left one. how do we check this? simple max value!

    assert np.argmax(most_left_spec) < np.argmax(most_right_spec)


def test_doppler_shift_grid_approach_side_more_left_neg_vrot(
    test_make_grid_values, test_load_phoenix_model
):
    grid = test_make_grid_values
    star_wave, star_flux = test_load_phoenix_model
    n_wl = len(star_wave)

    spectrum_grid = doppler_shift_grid(grid, star_flux, star_wave, -3)

    # from the approach side, the star should be more blue.
    max_r = np.max(grid[:, 0])
    most_right = grid[:, 1][np.argmin(np.abs(grid[:, 1] - 0))]
    most_left = grid[:, 1][np.argmin(np.abs(grid[:, 1] - np.pi))]

    # get the gridpoint that is most right and max r

    most_right_idx = np.argwhere((grid[:, 0] == max_r) & (grid[:, 1] == most_right))[0][
        0
    ]
    most_left_idx = np.argwhere((grid[:, 0] == max_r) & (grid[:, 1] == most_left))[0][0]

    # now get the spectra at these points from the spectrum grid.

    most_right_spec = spectrum_grid[most_right_idx]
    most_left_spec = spectrum_grid[most_left_idx]

    # now, the right one should be redshifted compared to the left one. how do we check this? simple max value!

    assert np.argmax(most_left_spec) > np.argmax(most_right_spec)


@pytest.fixture()
def test_doppler_shift_grid_baseline(test_make_grid_values, test_load_phoenix_model):
    grid = test_make_grid_values
    star_wave, star_flux = test_load_phoenix_model
    n_wl = len(star_wave)

    spectrum_grid = doppler_shift_grid(grid, star_flux, star_wave, 5000)
    return spectrum_grid


def test_average_spectrum_not_input(
    test_load_phoenix_model, test_doppler_shift_grid_baseline
):
    star_wave, star_flux = test_load_phoenix_model
    spectrum_grid = test_doppler_shift_grid_baseline
    assert not np.allclose(star_flux, np.mean(spectrum_grid, axis=0))


def test_occult_grid_shape(test_make_grid_values, test_doppler_shift_grid_baseline):
    grid = test_make_grid_values
    spectrum_grid = test_doppler_shift_grid_baseline
    occulted_grid = np.copy(spectrum_grid)
    planet_location = np.array([0, np.pi / 2])
    r_p = 0.1
    occulted_grid = occult_grid(occulted_grid, grid, planet_location, r_p)

    # what are the parts of the grid that are occulted when the planet is at the center of the star?

    assert np.allclose(occulted_grid.shape, spectrum_grid.shape)


def test_occult_grid_tiny_planet_does_not_matter(
    test_make_grid_values, test_doppler_shift_grid_baseline
):
    grid = test_make_grid_values
    spectrum_grid = test_doppler_shift_grid_baseline
    occulted_grid = np.copy(spectrum_grid)
    planet_location = np.array([0, np.pi / 2])
    r_p = 1e-16
    occulted_grid = occult_grid(occulted_grid, grid, planet_location, r_p)

    # what are the parts of the grid that are occulted when the planet is at the center of the star?

    assert np.allclose(occulted_grid, spectrum_grid)


def test_occult_grid_big_planet_far_away_does_not_matter(
    test_make_grid_values, test_doppler_shift_grid_baseline
):
    grid = test_make_grid_values
    spectrum_grid = test_doppler_shift_grid_baseline
    occulted_grid = np.copy(spectrum_grid)
    planet_location = np.array([50, np.pi / 2])
    r_p = 2
    occulted_grid = occult_grid(occulted_grid, grid, planet_location, r_p)

    # what are the parts of the grid that are occulted when the planet is at the center of the star?

    assert np.allclose(occulted_grid, spectrum_grid)


def test_occult_grid_star_planet_allgone(
    test_make_grid_values, test_doppler_shift_grid_baseline
):
    """
    if the planet is the size of the star, then the star is gone.
    """
    grid = test_make_grid_values
    spectrum_grid = test_doppler_shift_grid_baseline
    occulted_grid = np.copy(spectrum_grid)
    planet_location = np.array([0, np.pi / 2])
    r_p = 1.0
    occulted_grid = occult_grid(occulted_grid, grid, planet_location, r_p)

    # what are the parts of the grid that are occulted when the planet is at the center of the star?

    assert np.allclose(occulted_grid, np.zeros_like(occulted_grid))


def test_occult_grid_little_decrease(
    test_make_grid_values, test_doppler_shift_grid_baseline
):
    """
    there should be a decrease in the average flux during transit.
    """
    grid = test_make_grid_values
    spectrum_grid = test_doppler_shift_grid_baseline
    occulted_grid = np.copy(spectrum_grid)
    planet_location = np.array([0, np.pi / 2])
    r_p = 1.0
    occulted_grid = occult_grid(occulted_grid, grid, planet_location, r_p)

    # what are the parts of the grid that are occulted when the planet is at the center of the star?

    assert np.average(occulted_grid) < np.average(spectrum_grid)


def test_occult_correct_part_of_grid(
    test_make_grid_values, test_doppler_shift_grid_baseline
):
    """
    let's block off a bit of one side. rest should be normal.
    """
    grid = test_make_grid_values
    spectrum_grid = test_doppler_shift_grid_baseline
    occulted_grid = np.copy(spectrum_grid)
    planet_location = np.array([0.5, np.pi])
    r_p = 0.01
    occulted_grid = occult_grid(occulted_grid, grid, planet_location, r_p)

    mask = (grid[:, 1] < np.pi / 2) | (grid[:, 1] > 3 * np.pi / 2)
    occulted_grid_right = occulted_grid[mask]
    spectrum_grid_right = spectrum_grid[mask]

    # what are the parts of the grid that are occulted when the planet is at the center of the star?

    np.testing.assert_array_equal(occulted_grid_right, spectrum_grid_right)


def test_sum_grid_shape(test_make_grid_values, test_doppler_shift_grid_baseline):
    grid = test_make_grid_values
    spectrum_grid = test_doppler_shift_grid_baseline
    areas = np.ones(grid.shape[0])
    summed_grid = sum_grid(spectrum_grid, areas)

    assert summed_grid.shape == (len(spectrum_grid[0]),)


def test_sum_grid_equal_areas(test_make_grid_values, test_doppler_shift_grid_baseline):
    grid = test_make_grid_values
    spectrum_grid = test_doppler_shift_grid_baseline
    areas = np.ones(grid.shape[0])
    summed_grid = sum_grid(spectrum_grid, areas)
    assert np.allclose(summed_grid, np.mean(spectrum_grid, axis=0))


@pytest.fixture
def test_stellar_disk(test_load_phoenix_model):
    star_wave, star_flux = test_load_phoenix_model
    v_rot = 3000
    phases = np.linspace(-0.01, 0.01, 10)
    r_star = 1e9
    a = 3e9
    inc = np.pi / 2
    lambda_misalign = 0
    r_p = 1e8

    summed_grids, correction_factor, areas, occulted_grid = make_stellar_disk(
        star_flux,
        star_wave,
        v_rot,
        phases,
        r_star,
        inc,
        lambda_misalign,
        a,
        r_p,
        n_theta=10,
        n_r=10,
    )
    return summed_grids, correction_factor, phases, areas, occulted_grid


def test_make_stellar_disk_shape(test_stellar_disk, test_load_phoenix_model):
    star_wave, star_flux = test_load_phoenix_model
    summed_grids, correction_factor, phases, areas, occulted_grid = test_stellar_disk
    assert summed_grids.shape[0] == len(phases) and summed_grids.shape[1] == len(
        star_flux
    )


def test_make_stellar_disk_correction_factor_regular(
    test_stellar_disk, test_load_phoenix_model
):
    star_wave, star_flux = test_load_phoenix_model
    summed_grids, correction_factor, phases, areas, occulted_grid = test_stellar_disk
    assert np.all(np.isfinite(correction_factor)) and np.all(np.isfinite(summed_grids))


# def test_make_stellar_disk_summed_grids_factor_small(test_stellar_disk):
#     """
#     all the values in the summed grids should be less than 1.
#     """
#     summed_grids, correction_factor, phases, areas, occulted_grid = test_stellar_disk
#     assert np.all(summed_grids < 1.)


def test_make_stellar_disk_bigger_planet_more_occulted(
    test_stellar_disk, test_load_phoenix_model
):
    """
    if it's a bigger planet, more should be occulted.
    """
    summed_grids, correction_factor, phases, areas, occulted_grid = test_stellar_disk
    star_wave, star_flux = test_load_phoenix_model
    v_rot = 3000
    phases = np.linspace(-0.01, 0.01, 10)
    r_star = 1e9
    a = 3e9
    inc = np.pi / 2
    lambda_misalign = 0
    r_p = 5e8

    new_summed_grids, new_correction_factor, areas, occulted_grid = make_stellar_disk(
        star_flux, star_wave, v_rot, phases, r_star, inc, lambda_misalign, a, r_p
    )
    assert np.all(new_summed_grids <= summed_grids)



"""
Stores constants used in the scope package. We store the values themselves so that jax and numba need not have access
to astropy or scipy.
"""
import astropy.constants as const

const_c = const.c.si.value
rjup_rsun = (const.R_jup / const.R_sun).si.value
rsun = const.R_sun.si.value



# import sys,os
# sys.path.append(os.path.realpath('..'))
# sys.path.append(os.path.realpath('.'))
import os
import sys

import jax
import numpy as np
from tqdm import tqdm

from scope.broadening import *
from scope.ccf import *
from scope.input_output import parse_input_file, write_input_file
from scope.noise import *
from scope.tellurics import *
from scope.utils import *

abs_path = os.path.dirname(__file__)


def make_data(
    scale,
    wlgrid,
    wl_model,
    Fp_conv,
    Fstar_conv,
    n_order,
    n_exposure,
    n_pixel,
    phases,
    Rp_solar,
    Rstar,
    seed=42,
    do_pca=True,
    blaze=False,
    tellurics=False,
    n_princ_comp=4,
    v_sys=0.0,
    Kp=192.06,
    star=False,
    SNR=0,
    rv_semiamp_orbit=0.3229,
    observation="emission",
    tell_type="ATRAN",
    time_dep_tell=False,
    wav_error=False,
    order_dep_throughput=False,
    a=1,  # AU
    u1=0.3,
    u2=0.3,
    LD=True,
    b=0.0,  # impact parameter
    divide_out_of_transit=False,
    out_of_transit_dur=0.1,
    v_sys_measured=0.0,
):
    """
    Creates a simulated HRCCS dataset. Main function.

    Inputs
    ------
        :scale: (float) scaling factor for the data.
        :wlgrid: (array) wavelength grid for the data.
        :do_pca: (bool) if True, do PCA on the data.
        :blaze: (bool) if True, include the blaze function in the data.
        :tellurics: (bool) if True, include tellurics in the data.
        :n_princ_comp: (int) number of principal components to use.
        :v_sys: (float) systemic velocity of the planet.
        :Kp: (float) semi-amplitude of the planet.
        :star: (bool) if True, include the stellar spectrum in the data.
        :SNR: (float) photon signal-to-noise ratio (not of the planetary spectrum itself!)
        :observation: (str) 'emission' or 'transmission'.
        :tell_type: (str) 'ATRAN' or 'data-driven'.
        :time_dep_tell: (bool) if True, include time-dependent tellurics. not applicable for ATRAN.
        :wav_error: (bool) if True, include wavelength-dependent error.
        :order_dep_throughput: (bool) if True, include order-dependent throughput.

    Outputs
    -------
        :pca_noise_matrix: (array) the data cube with only the larger-scale trends.
        :flux_cube: (array) the data cube with the larger-scale trends removed.
        :fTemp_nopca: (array) the data cube with all components.
        :just_tellurics: (array) the telluric model that's multiplied to the dataset.

    """
    np.random.seed(seed)

    rv_planet, rv_star = calc_rvs(
        v_sys, v_sys_measured, Kp, rv_semiamp_orbit, phases
    )  # measured in m/s now

    flux_cube = np.zeros(
        (n_order, n_exposure, n_pixel)
    )  # will store planet and star signal
    pca_noise_matrix = np.zeros((n_order, n_exposure, n_pixel))
    for order in range(n_order):
        wlgrid_order = np.copy(wlgrid[order,])  # Cropped wavelengths
        flux_cube[order] = doppler_shift_planet_star(
            flux_cube[order],
            n_exposure,
            phases,
            rv_planet,
            rv_star,
            wlgrid_order,
            wl_model,
            Fp_conv,
            Rp_solar,
            Fstar_conv,
            Rstar,
            u1,
            u2,
            a,
            b,
            LD,
            scale,
            star,
            observation,
        )

    throughput_baselines = np.loadtxt(abs_path + "/data/throughputs.txt")

    flux_cube = detrend_cube(flux_cube, n_order, n_exposure)

    for i in range(flux_cube.shape[1]):
        throughput_baseline = throughput_baselines[i]
        throughput_factor = throughput_baseline / np.nanpercentile(
            flux_cube[:, i, :], 90
        )
        flux_cube[:, i, :] *= throughput_factor

    if tellurics:
        flux_cube = add_tellurics(
            wlgrid,
            flux_cube,
            n_order,
            n_exposure,
            vary_airmass=True,
            tell_type=tell_type,
            time_dep_tell=time_dep_tell,
        )
        # these spline fits aren't perfect. we mask negative values to 0.
        just_tellurics = np.ones_like(flux_cube)
        just_tellurics = add_tellurics(
            wlgrid,
            just_tellurics,
            n_order,
            n_exposure,
            vary_airmass=True,
            tell_type=tell_type,
            time_dep_tell=time_dep_tell,
        )
        flux_cube[flux_cube < 0.0] = 0.0
        just_tellurics[just_tellurics < 0.0] = 0.0
    flux_cube = detrend_cube(flux_cube, n_order, n_exposure)
    if blaze:
        flux_cube = add_blaze_function(wlgrid, flux_cube, n_order, n_exposure)
        flux_cube[flux_cube < 0.0] = 0.0
    flux_cube[np.isnan(flux_cube)] = 0.0

    flux_cube = detrend_cube(flux_cube, n_order, n_exposure)
    flux_cube[np.isnan(flux_cube)] = 0.0
    if SNR > 0:  # 0 means don't add noise!
        if order_dep_throughput:
            noise_model = "IGRINS"
        else:
            noise_model = "constant"
        flux_cube = add_noise_cube(flux_cube, wlgrid, SNR, noise_model=noise_model)

    flux_cube = detrend_cube(flux_cube, n_order, n_exposure)

    if wav_error:
        doppler_shifts = np.loadtxt(
            "data/doppler_shifts_w77ab.txt"
        )  # todo: create this!
        flux_cube = change_wavelength_solution(wlgrid, flux_cube, doppler_shifts)

    flux_cube = detrend_cube(flux_cube, n_order, n_exposure)
    flux_cube[np.isnan(flux_cube)] = 0.0
    flux_cube_nopca = flux_cube.copy()

    if observation == "transmission" and divide_out_of_transit:
        # generate the out of transit baseline
        n_exposures_baseline = (
            out_of_transit_dur * n_exposure
        )  # assuming n_exposure is fully just in transit
        out_of_transit_flux = np.ones_like(
            flux_cube
        )  # todo: fix shape for out of transit baseline

        # take star
        for exposure in range(n_exposures_baseline):
            flux_star = calc_doppler_shift(
                wlgrid_order, wl_model, Fstar_conv, rv_star[exposure]
            )

            out_of_transit_flux[exposure,] *= flux_star

        out_of_transit_flux = detrend_cube(out_of_transit_flux, n_order, n_exposure)

        # add tellurics
        out_of_transit_flux = add_tellurics(
            wlgrid,
            out_of_transit_flux,
            n_order,
            n_exposures_baseline,
            vary_airmass=True,
            tell_type=tell_type,
            time_dep_tell=time_dep_tell,
        )
        out_of_transit_flux = detrend_cube(out_of_transit_flux, n_order, n_exposure)

        # add blaze
        out_of_transit_flux = add_blaze_function(
            wlgrid, out_of_transit_flux, n_order, n_exposures_baseline
        )
        out_of_transit_flux[flux_cube < 0.0] = 0.0
        out_of_transit_flux[np.isnan(flux_cube)] = 0.0
        out_of_transit_flux = detrend_cube(out_of_transit_flux, n_order, n_exposure)

        # add noise
        out_of_transit_flux = add_noise_cube(
            out_of_transit_flux, wlgrid, SNR, noise_model=noise_model
        )

        # take median
        median_out_of_transit = np.median(out_of_transit_flux, axis=1)

        # divide out the flux cube
        flux_cube /= median_out_of_transit  # todo: check axes work out

    if do_pca:
        for j in range(n_order):
            flux_cube[j] -= np.mean(flux_cube[j])
            flux_cube[j] /= np.std(flux_cube[j])
            flux_cube[j], pca_noise_matrix[j] = perform_pca(
                flux_cube[j], n_princ_comp, return_noplanet=True
            )
            # todo: think about the svd
            # todo: redo all analysis centering on 0?
    else:
        for j in range(n_order):
            for i in range(n_exposure):
                flux_cube[j][i] -= np.mean(flux_cube[j][i])

    if np.all(pca_noise_matrix == 0):
        print("was all zero")
        pca_noise_matrix = np.ones_like(pca_noise_matrix)
    if tellurics:
        return pca_noise_matrix, flux_cube, flux_cube_nopca, just_tellurics
    return (
        pca_noise_matrix,
        flux_cube,
        flux_cube_nopca,
        np.ones_like(flux_cube),
    )  # returning CCF and logL values


# @njit
def calc_log_likelihood(
    v_sys,
    Kp,
    scale,
    wlgrid,
    wl_model,
    Fp_conv,
    Fstar_conv,
    flux_cube,
    n_order,
    n_exposure,
    n_pixel,
    phases,
    Rp_solar,
    Rstar,
    rv_semiamp_orbit,
    A_noplanet,
    do_pca=True,
    n_princ_comp=4,
    star=False,
    observation="emission",
    a=1,  # AU
    u1=0.3,
    u2=0.3,
    LD=True,
    b=0.0,  # impact parameter
    v_sys_measured=0.0,
):
    """
    Calculates the log likelihood and cross-correlation function of the data given the model parameters.

    Inputs
    ------
          :v_sys: float
            Systemic velocity of the system in m/s
            :Kp: float
            Planet semi-amplitude in m/s
            :scale: float
            Planet-to-star flux ratio
            :wlgrid: array
            Wavelength grid
            :Fp_conv: array
            Planet spectrum convolved with the instrumental profile
            :Fstar_conv: array
            Stellar spectrum convolved with the instrumental profile
            :flux_cube: array
            Data cube
            :n_order: int
            Number of orders
            :n_exposure: int
            Number of exposures
            :n_pixel: int
            Number of pixels
            :phases: array
            Orbital phases
            :Rp_solar: float
            Planet radius in solar radii
            :Rstar: float
            Stellar radius in solar radii
            :rv_semiamp_orbit: float
            Semi-amplitude of the orbit in km/s. This is the motion of the *star*.
            :A_noplanet: array
            Array of the non-planet component of the data cube
            :do_pca: bool
            Whether to perform PCA on the data cube
            :n_princ_comp: int
            Number of principal components to use in the PCA
            :star: bool
            Whether to include the stellar component in the simulation.
            :observation: str
            Type of observation. Currently supported: 'emission', 'transmission'
    Outputs
    -------
            :logL: float
            Log likelihood of the data given the model parameters
            :ccf: array
            Cross-correlation function of the data given the model parameters
    """
    rv_planet, rv_star = calc_rvs(
        v_sys, v_sys_measured, Kp, rv_semiamp_orbit, phases
    )  # measured in m/s
    CCF = 0.0
    logL = 0.0
    for order in range(n_order):
        wlgrid_order = np.copy(wlgrid[order,])  # Cropped wavelengths
        model_flux_cube = np.zeros(
            (n_exposure, n_pixel)
        )  # "shifted" model spectra array at each phase

        model_flux_cube = doppler_shift_planet_star(
            model_flux_cube,
            n_exposure,
            phases,
            rv_planet,
            rv_star,
            wlgrid_order,
            wl_model,
            Fp_conv,
            Rp_solar,
            Fstar_conv,
            Rstar,
            u1,
            u2,
            a,
            b,
            LD,
            scale,
            star,
            observation,
            reprocessing=True,
        )

        # ok now do the PCA. where does it fall apart?
        if do_pca:
            # process the model same as the "data"!
            model_flux_cube *= A_noplanet[order]
            model_flux_cube, _ = perform_pca(model_flux_cube, n_princ_comp, False)
        # I = np.ones(n_pixel)

        logl, ccf = calc_ccf(model_flux_cube, flux_cube[order], n_pixel)
        CCF += ccf
        logL += logl

        # # todo: airmass detrending reprocessing

    return logL, CCF  # returning CCF and logL values


def simulate_observation(
    planet_spectrum_path=".",
    star_spectrum_path=".",
    data_cube_path=".",
    phase_start=0,
    phase_end=1,
    n_exposures=10,
    observation="emission",
    blaze=True,
    n_princ_comp=4,
    star=True,
    SNR=250,
    telluric=True,
    tell_type="data-driven",
    time_dep_tell=False,
    wav_error=False,
    rv_semiamp_orbit=0.3229,
    order_dep_throughput=True,
    Rp=1.21,  # Jupiter radii,
    Rstar=0.955,  # solar radii
    kp=192.02,  # planetary orbital velocity, km/s
    v_rot=4.5,
    scale=1.0,
    v_sys=0.0,
    modelname="yourfirstsimulation",
    divide_out_of_transit=False,
    out_of_transit_dur=0.1,
    include_rm=False,
    v_rot_star=3.0,
    a=0.033,  #
    lambda_misalign=0.0,
    inc=90.0,
    seed=42,
    **kwargs,
):
    """
    Run a simulation of the data, given a grid index and some paths. Side effects:
    writes some files in the output directory.

    Inputs
    ------
    grid_ind: int
        The index of the grid to use.
    planet_spectrum_path: str
        The path to the planet spectrum.
    star_spectrum_path: str
        The path to the star spectrum.
    phases: array-like
        The phases of the observations.
    observation: str
        Type of observation to simulate. Currently supported: emission, transmission.

    Outputs
    -------
    None

    """
    # make the output directory

    outdir = abs_path + f"/output/{modelname}"

    make_outdir(outdir)

    # and write the input file out
    output_args = locals()
    output_args.update(kwargs)
    write_input_file(output_args, output_file_path=f"{outdir}/input.txt")

    phases = np.linspace(phase_start, phase_end, n_exposures)
    Rp_solar = Rp * rjup_rsun  # convert from jupiter radii to solar radii
    Kp_array = np.linspace(kp - 100, kp + 100, 200)
    v_sys_array = np.arange(v_sys - 100, v_sys + 100)
    n_order, n_pixel = (44, 1848)  # todo: generalize.
    mike_wave, mike_cube = pickle.load(open(data_cube_path, "rb"), encoding="latin1")

    wl_cube_model = mike_wave.copy().astype(np.float64)

    wl_model, Fp, Fstar = np.load(planet_spectrum_path, allow_pickle=True)

    wl_model = wl_model.astype(np.float64)

    # Fp_conv_rot = broaden_spectrum(wl_model / 1e6, Fp, 0, vl=v_rot)
    rot_ker = get_rot_ker(v_rot, wl_model)
    Fp_conv_rot = np.convolve(Fp, rot_ker, mode="same")

    # instrument profile convolution
    instrument_kernel = get_instrument_kernel()
    Fp_conv = np.convolve(Fp_conv_rot, instrument_kernel, mode="same")

    star_wave, star_flux = np.loadtxt(
        star_spectrum_path
    ).T  # Phoenix stellar model packing

    if include_rm:
        star_flux, _ = make_stellar_disk(
            star_flux, star_wave, v_rot_star, phases, Rstar, inc, lambda_misalign, a, Rp
        )

    Fstar_conv = get_star_spline(
        star_wave, star_flux, wl_model, instrument_kernel, smooth=False
    )

    lls, ccfs = np.zeros((200, 200)), np.zeros((200, 200))

    # redoing the grid. how close does PCA get to a tellurics-free signal detection?
    A_noplanet, flux_cube, flux_cube_nopca, just_tellurics = make_data(
        scale,
        wl_cube_model,
        wl_model,
        Fp_conv,
        Fstar_conv,
        n_order,
        n_exposures,
        n_pixel,
        phases,
        Rp_solar,
        Rstar,
        seed=seed,
        do_pca=True,
        blaze=blaze,
        n_princ_comp=n_princ_comp,
        tellurics=telluric,
        v_sys=v_sys,
        star=star,
        Kp=kp,
        SNR=SNR,
        rv_semiamp_orbit=rv_semiamp_orbit,
        tell_type=tell_type,
        time_dep_tell=time_dep_tell,
        wav_error=wav_error,
        order_dep_throughput=order_dep_throughput,
        observation=observation,
        divide_out_of_transit=False,
        out_of_transit_dur=0.1,
        v_sys_measured=v_sys,
    )

    run_name = f"{n_princ_comp}_NPC_{blaze}_blaze_{star}_star_{telluric}_telluric_{SNR}_SNR_{tell_type}_{time_dep_tell}_{wav_error}_{order_dep_throughput}"

    save_data(outdir, run_name, flux_cube, flux_cube_nopca, A_noplanet, just_tellurics)

    for l, Kp in tqdm(
        enumerate(Kp_array), total=len(Kp_array), desc="looping PCA over Kp"
    ):
        for k, v_sys in enumerate(v_sys_array):
            res = calc_log_likelihood(
                v_sys,
                Kp,
                scale,
                wl_cube_model,
                wl_model,
                Fp_conv,
                Fstar_conv,
                flux_cube,
                n_order,
                n_exposures,
                n_pixel,
                phases,
                Rp_solar,
                Rstar,
                rv_semiamp_orbit,
                do_pca=True,
                n_princ_comp=n_princ_comp,
                A_noplanet=A_noplanet,
                star=star,
                observation=observation,
                v_sys_measured=v_sys,
            )
            lls[l, k], ccfs[l, k] = res

    save_results(outdir, run_name, lls, ccfs)


if __name__ == "__main__":
    file = "input.txt"
    inputs = parse_input_file(file)
    simulate_observation(**inputs)



import unittest

import numpy as np

import pytest

from scope.utils import *


class Testdetrend_cube(unittest.TestCase):
    test_cube = np.random.random((10, 10, 10))

    def test_cube_normalized(self):
        n_order, n_exp, n_pix = self.test_cube.shape
        detrended_cube = detrend_cube(self.test_cube, n_order, n_exp)
        self.assertTrue(np.max(detrended_cube) == 1)

    def test_cube_normalized_same_shape(self):
        original_shape = self.test_cube.shape
        n_order, n_exp, n_pix = original_shape
        detrended_cube = detrend_cube(self.test_cube, n_order, n_exp)
        self.assertTrue(detrended_cube.shape == original_shape)

    def test_cube_normalized_min_still_min(self):
        orig_min_arr = np.argmin(self.test_cube, axis=0)
        n_order, n_exp, n_pix = self.test_cube.shape
        detrended_cube = detrend_cube(self.test_cube, n_order, n_exp)
        self.assertTrue(np.all(orig_min_arr == np.argmin(detrended_cube, axis=0)))

    def test_cube_normalized_max_still_max(self):
        orig_min_arr = np.argmax(self.test_cube, axis=0)
        n_order, n_exp, n_pix = self.test_cube.shape
        detrended_cube = detrend_cube(self.test_cube, n_order, n_exp)
        self.assertTrue(np.all(orig_min_arr == np.argmax(detrended_cube, axis=0)))


# todo: benchmark against batman at some point.


@pytest.mark.parametrize(
    "values, output",
    [
        ([0, 0, 1, 0, 1, 0, True], 1.0),  # if no coefficients, no LD
        ([0, 0, 1, 0, 1, 0, False], 1.0),  # but that doesn't matter if LD is turned off
        (
            [1, 1, 10 * rsun, 0, 1e-3, 0.3, True],
            0.0,
        ),  # if there is a really tiny sun, you're not gonna hit the sun
        ([1, 1, 1, 0, 1, 0, True], 1.0),  # at center phase, should be 1
        (
            [0.3, 0.3, 10 * rsun, 0.0, 1, 0.25, True],
            0.0,
        ),  # at quadrature you're definitely not transiting!
        (
            [0.3, 0.3, 10 * rsun, 1.3, 1, 0.0, True],
            0.0,
        ),  # non-transiting planets don't transit!
    ],
)
def test_calc_limb_darkening(values, output):
    assert calc_limb_darkening(*values) == output


class TestPca(unittest.TestCase):
    """
    These don't really test the fancy statistics of PCA. just assert that it's a
    variance-reducing black box.
    """

    n_exp = 10
    n_pix = 10
    cube = np.random.random((n_exp, n_pix))

    def test_perform_pca_shapes(self):
        """
        just make sure the output shapes are what we expect.
        """

        n_princ_comp = 5
        scaled_cube, pca = perform_pca(self.cube, n_princ_comp, return_noplanet=True)
        assert scaled_cube.shape == self.cube.shape
        assert pca.shape == self.cube.shape

    def test_perform_pca_variance_removed(self):
        """
        just make sure the output shapes are what we expect.
        """

        n_princ_comp = 5
        scaled_cube, pca = perform_pca(self.cube, n_princ_comp, return_noplanet=True)
        assert np.std(scaled_cube) < np.std(
            pca
        )  # the variance should be contained in the main components

    def test_perform_pca_variance_removed_compared_input(self):
        """
        just make sure the output shapes are what we expect.
        """

        n_princ_comp = 5
        scaled_cube, pca = perform_pca(self.cube, n_princ_comp, return_noplanet=True)
        assert np.std(scaled_cube) < np.std(
            self.cube
        )  # the variance should be contained in the main components

    def test_perform_pca_variance_removed_compared_input_diffcompoents(self):
        """
        just make sure the output shapes are what we expect.
        """

        n_princ_comp = 5
        scaled_cube5, pca5 = perform_pca(self.cube, n_princ_comp, return_noplanet=True)
        n_princ_comp = 10
        scaled_cube10, pca10 = perform_pca(
            self.cube, n_princ_comp, return_noplanet=True
        )
        assert np.std(scaled_cube10) < np.std(
            scaled_cube5
        )  # the variance should be contained in the main components


class TestDopplerShift(unittest.TestCase):
    def test_same_after_shift(self):
        """
        shouldn't globally change the properties. should really just shift things.
        """
        eval_wave = np.linspace(1.1, 1.8, 10000) * 1e-6  # meters.
        template_wave = np.linspace(1, 2, 10000) * 1e-6

        template_flux = np.random.random(10000)
        interped_flux = np.interp(eval_wave, template_wave, template_flux)
        v = 1e-6  # m/s. should not change much

        shifted_flux = calc_doppler_shift(eval_wave, template_wave, template_flux, v)
        # test that shifted flux and interpred flux are very similar
        assert (
            np.testing.assert_allclose(shifted_flux, interped_flux, rtol=1e-2) == None
        )

    def test_diff_after_shift(self):
        """
        shouldn't globally change the properties. should really just shift things.
        """
        eval_wave = np.linspace(1.1, 1.8, 10000) * 1e-6  # meters.
        template_wave = np.linspace(1, 2, 10000) * 1e-6

        template_flux = np.random.random(10000)
        interped_flux = np.interp(eval_wave, template_wave, template_flux)
        v = 2e3  # m/s. should not change much

        shifted_flux = calc_doppler_shift(eval_wave, template_wave, template_flux, v)
        # test that shifted flux and interpred flux are very similar
        np.testing.assert_raises(
            AssertionError, np.testing.assert_array_equal, shifted_flux, interped_flux
        )

    def test_gaussian_shifted_blue(self):
        """
        shouldn't globally change the properties. should really just shift things.
        """
        eval_wave = np.linspace(1.1, 1.8, 10000) * 1e-6
        template_wave = np.linspace(1, 2, 10000) * 1e-6
        # make a gaussian centered at 1.5
        template_flux = np.exp(-0.5 * ((template_wave - 1.5e-6) / 0.01) ** 2)
        interped_flux = np.interp(eval_wave, template_wave, template_flux)

        v = -5e4  # m/s. should not change much
        shifted_flux = calc_doppler_shift(eval_wave, template_wave, template_flux, v)
        wav_max_shifted = eval_wave[np.argmax(shifted_flux)]
        wav_max = eval_wave[np.argmax(interped_flux)]
        assert wav_max_shifted < wav_max

    def test_gaussian_shifted_red(self):
        """
        shouldn't globally change the properties. should really just shift things.
        """
        eval_wave = np.linspace(1.1, 1.8, 10000) * 1e-6
        template_wave = np.linspace(1, 2, 10000) * 1e-6
        # make a gaussian centered at 1.5
        template_flux = np.exp(-0.5 * ((template_wave - 1.5e-6) / 0.01) ** 2)

        v = 5e4  # m/s. should not change much
        shifted_flux = calc_doppler_shift(eval_wave, template_wave, template_flux, v)
        interped_flux = np.interp(eval_wave, template_wave, template_flux)
        wav_max_shifted = eval_wave[np.argmax(shifted_flux)]
        wav_max = eval_wave[np.argmax(interped_flux)]
        assert wav_max_shifted > wav_max


class TestCrossingTime(unittest.TestCase):
    """
    tests the exposure time calculation.
    """

    period = 1.80988198
    mstar = 1.458
    e = 0.000
    inc = 89.623
    mplanet = 0.894
    rstar = 1.756
    peri = 0
    b = 0.027
    R = 45000
    pix_per_res = 3.3
    plot = False

    def test_crossing_time_moremassive_star(self):
        """
        in a greater potential well, the crossing time is shorter.
        """
        low_mass_time, _, _, _ = calc_crossing_time()

        high_mass_time, _, _, _ = calc_crossing_time(mstar=10 * self.mstar)
        assert low_mass_time > high_mass_time

    def test_crossing_time_moremassive_star(self):
        """
        on a longer period for the same star, the crossing time is longer.
        """
        short_period_time, _, _, _ = calc_crossing_time()

        long_period_time, _, _, _ = calc_crossing_time(period=10 * self.period)
        assert short_period_time < long_period_time

    def test_crossing_time_pixel_per_res(self):
        """
        more pixels per resolution element, shorter time to cross the pixels.
        """
        _, few_pixels_per_element, _, _ = calc_crossing_time()

        _, many_pixels_per_element, _, _ = calc_crossing_time(
            pix_per_res=10 * self.pix_per_res
        )
        assert many_pixels_per_element < few_pixels_per_element

    def test_crossing_time_resolution(self):
        """
        higher resolution, shorter crossing time.
        """
        low_resolution_time, _, _, _ = calc_crossing_time()

        high_resolution_time, _, _, _ = calc_crossing_time(R=10 * self.R)
        assert high_resolution_time < low_resolution_time



import io
import os
from datetime import datetime

import pandas as pd

from scope.calc_quantities import *


class ScopeConfigError(Exception):
    def __init__(self, message="scope input file error:"):
        self.message = message
        super().__init__(self.message)


# Mapping between input file parameters and database columns
parameter_mapping = {
    "Rp": "pl_radj",
    "Rstar": "st_rad",
    "v_sys": "system_velocity",
    "a": "pl_orbsmax",
    "P_rot": "pl_orbper",
    "v_sys": "st_radv",
    "planet_name": "pl_name",
    "Rp_solar": "planet_radius_solar",
}


def query_database(
    planet_name, parameter, database_path="data/default_params_exoplanet_archive.csv"
):
    if not os.path.exists(database_path):
        raise FileNotFoundError(f"Database file {database_path} not found.")

    df = pd.read_csv(database_path)

    # check whether planet name is in database
    if planet_name not in df["pl_name"].values:
        print(f"Planet name {planet_name} not found in database.")
    try:
        # Use the mapped parameter name if it exists, otherwise use the original
        db_parameter = parameter_mapping.get(parameter, parameter)
        value = df.loc[df["pl_name"] == planet_name, db_parameter].values[0]
        return float(value)
    except Exception as e:
        print(f"Error querying database for {planet_name}, {parameter}: {e}")

        return np.nan


def unpack_lines(content):
    """
    Unpack lines from a file, removing comments and joining lines that are split.

    Parameters
    ----------
    content : list
        List of lines from the file.

    Returns
    -------
    data_lines : list
        List of lines with comments removed and split lines joined.
    """
    data_lines = []
    for line in content:
        line = line.strip()
        if line.startswith("Planet name:"):
            planet_name = line.split("Planet name:", 1)[1].strip()
        elif "Author:" in line:
            author = line.split("Author:", 1)[1].strip()
        elif (
            not line.startswith("#")
            and not line.startswith(":")
            and not line.startswith("Created")
            and line
        ):
            data_lines.append(line)
    return data_lines, planet_name, author


def coerce_nulls(data, key, value):
    if value == "NULL":
        data[key] = np.nan

    return data


def coerce_integers(data, key, value):
    integer_fields = ["n_exposures", "n_princ_comp"]
    if key in integer_fields:
        data[key] = int(value)
    else:
        try:
            data[key] = float(value)
        except:
            pass

    return data


def coerce_database(data, key, value, astrophysical_params, planet_name, database_path):
    if value == "DATABASE" and key in astrophysical_params:
        data[key] = query_database(planet_name, key, database_path)
    elif value == "DATABASE" and key in ["phase_start", "phase_end"]:
        tdur = query_database(planet_name, "pl_trandur", database_path)
        period = query_database(planet_name, "pl_orbper", database_path)

        # convert it to phase
        tdur_phase = convert_tdur_to_phase(tdur, period)

        if key == "phase_start":
            data[key] = -tdur_phase / 2
        else:
            data[key] = tdur_phase / 2

    return data


def coerce_splits(data, key, value):
    if isinstance(value, str) and "," in value:
        data[key] = [float(v.strip()) for v in value.split(",")]

    return data


def coerce_booleans(data, key, value):
    if value == "True":
        data[key] = True
    elif value == "False":
        data[key] = False

    return data


def parse_input_file(
    file_path, database_path="data/default_params_exoplanet_archive.csv", **kwargs
):
    """
    Parse an input file and return a dictionary of parameters.

    Parameters
    ----------
    file_path : str
        Path to the input file.
    database_path : str
        Path to the database file.
    **kwargs
        Additional keyword arguments to add to the data dictionary.

    Returns
    -------
    data : dict
        Dictionary of parameters.
    """
    # First, read the entire file content
    with open(file_path, "r") as file:
        content = file.readlines()

    data_lines, planet_name, author = unpack_lines(content)

    # Read the remaining lines with pandas
    df = pd.read_csv(
        io.StringIO("\n".join(data_lines)),
        sep=r"\s+",
        header=None,
        names=["parameter", "value"],
        comment="#",
    )

    # Convert the dataframe to a dictionary
    data = dict(zip(df["parameter"], df["value"]))

    # Add planet_name and author to the data dictionary
    data["planet_name"] = planet_name
    data["author"] = author

    # List of astrophysical parameters
    astrophysical_params = [
        "Rp",
        "Rp_solar",
        "Rstar",
        "kp",
        "v_rot",
        "v_sys",
        "P_rot",
        "a",
        "u1",
        "u2",
    ]

    # Convert values to appropriate types
    for key, value in data.items():
        # check for values that are comma-delimited
        data = coerce_splits(data, key, value)

        # Check for NULL
        data = coerce_nulls(data, key, value)

        # make sure integers are cast as such
        data = coerce_integers(data, key, value)

        # make sure booleans are cast as such
        data = coerce_booleans(data, key, value)

        # Check for DATABASE in astrophysical parameters
        data = coerce_database(
            data, key, value, astrophysical_params, planet_name, database_path
        )

    # Add any additional kwargs to the data dictionary
    data.update(kwargs)

    data = calculate_derived_parameters(data)

    if data["tell_type"] == "data-driven" and data["blaze"] == False:
        raise ScopeConfigError("Data-driven tellurics requires blaze set to True.")

    return data


def write_input_file(data, output_file_path="input.txt"):
    # Define the order and categories of parameters

    categories = {
        "Filepaths": ["planet_spectrum_path", "star_spectrum_path", "data_cube_path"],
        "Astrophysical Parameters": ["Rp", "Rp_solar", "Rstar", "kp", "v_rot", "v_sys"],
        "Instrument Parameters": ["SNR"],
        "Observation Parameters": [
            "observation",
            "phase_start",
            "phase_end",
            "n_exposures",
            "blaze",
            "star",
            "telluric",
            "tell_type",
            "time_dep_tell",
            "wav_error",
            "order_dep_throughput",
        ],
        "Analysis Parameters": ["n_princ_comp", "scale"],
    }
    current_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    with open(output_file_path, "w") as f:
        # Write the header
        f.write(
            f""":·········································
:                                       :
:    ▄▄▄▄▄   ▄█▄    ████▄ █ ▄▄  ▄███▄   :
:   █     ▀▄ █▀ ▀▄  █   █ █   █ █▀   ▀  :
: ▄  ▀▀▀▀▄   █   ▀  █   █ █▀▀▀  ██▄▄    :
:  ▀▄▄▄▄▀    █▄  ▄▀ ▀████ █     █▄   ▄▀ :
:            ▀███▀         █    ▀███▀   :
:                           ▀           :
:                                       :
:········································
Created: {current_date}
Author: YourName
Planet name: {data['planet_name']}

""".format(
                date=pd.Timestamp.now().strftime("%Y-%m-%d")
            )
        )
        # Write parameters by category
        for category, params in categories.items():
            f.write(f"# {category}\n")
            for param in params:
                if param in data:
                    value = data[param]
                    # Handle different types of values
                    if isinstance(value, bool):
                        value = str(value)
                    elif isinstance(value, float):
                        if np.isnan(value):
                            value = "NULL"
                        else:
                            value = f"{value:.6f}".rstrip("0").rstrip(".")
                    elif isinstance(value, list):
                        value = ",".join(map(str, value))
                    f.write(f"{param:<23} {value}\n")
            f.write("\n")

    print(f"Input file written to {output_file_path}")



"""
goals: make sure that grid can be indexed, no duplicates, etc.
"""

import unittest

from scope.grid import *


class TestGrid(unittest.TestCase):
    def test_indexing(self):
        dictionary = parameter_list[0]
        self.assertTrue(type(dictionary) == dict)

    def test_no_duplicates(self):
        new_list = []
        for dictionary in parameter_list:
            if dictionary not in new_list:
                new_list += [dictionary]
        self.assertTrue(len(new_list) == len(parameter_list))



import numpy as np
from sklearn.model_selection import ParameterGrid

blazes = [False, True]
order_dep_throughputs = [False]
tellurics = [False, True]
telluric_types = ["ATRAN"]
time_dep_tellurics = [False]
wav_errors = [False]
stars = [False, True]

NPCs = np.arange(5)

SNRs = [0, 60, 250]

parameter_list1 = list(
    ParameterGrid(
        {
            "blaze": blazes,
            "n_princ_comp": NPCs,
            "SNR": SNRs,
            "star": stars,
            "telluric": tellurics,
            "wav_error": wav_errors,
            "time_dep_telluric": time_dep_tellurics,
            "telluric_type": telluric_types,
            "order_dep_throughput": order_dep_throughputs,
        }
    )
)

"""
second grid — needs to include the new parameters.
    - new tellurics
    - new tellurics with time
    - wavelength-dependent throughput
    - Doppler shift jitter.

just make sure that I start indexing at the new one :)
"""


blazes = [True]

order_dep_throughputs = [True, False]

tellurics = [False, True]
telluric_types = ["data-driven"]
time_dep_tellurics = [False, True]
stars = [False, True]
wav_errors = [False, True]

NPCs = np.arange(5)

SNRs = [0, 60, 250]

parameter_list2 = list(
    ParameterGrid(
        {
            "blaze": blazes,
            "n_princ_comp": NPCs,
            "SNR": SNRs,
            "star": stars,
            "telluric": tellurics,
            "wav_error": wav_errors,
            "time_dep_telluric": time_dep_tellurics,
            "telluric_type": telluric_types,
            "order_dep_throughput": order_dep_throughputs,
        }
    )
)

parameter_list = parameter_list1 + parameter_list2



import matplotlib.pyplot as plt
import numpy as np
from astropy import constants as const
from numba import njit
from tqdm import tqdm

const_c = const.c.to("km/s").value


# @njit
def make_grid(n_r, n_theta):
    # well we don't want the edges of the grid, we want the midpoint.

    dr = 1 / n_r

    r_range = np.linspace(dr / 2, 1, n_r)  # in stellar radius units

    theta_range = np.linspace(0, 2 * np.pi, n_theta)  # in radians

    # fine just make it uniform theta grid and weight later on by dA.
    r_vals = np.tile(r_range, n_theta)
    theta_vals = np.repeat(theta_range, n_r)
    # define a new theta at each radius

    grid = np.zeros((n_theta * n_r, 2))  # not quite sure this is correct. return!
    grid[:, 0] = r_vals
    grid[:, 1] = theta_vals
    return grid


# @njit
def doppler_shift_grid(grid, flux, wavelengths, v_rot):
    """
    the grid is the (r, theta) values. takes in a 1D spectrum.

    v_rot is in meters per second.

    outputs the spectrum grid. each row is a different point on the grid. each column is a different wavelength.
    """
    # yeah this is where things get funky lol
    spectrum_grid = np.zeros((grid.shape[0], len(wavelengths)))

    # for i, point in tqdm(enumerate(grid), desc='doppler shifting', total=grid.shape[0]):
    for i, point in enumerate(grid):
        theta, r = point
        x = r * np.cos(theta)

        # calculate the v_dopp at that point assuming rotational broadening
        v_dopp = (
            x * v_rot
        )  # x is normalized from 0 to 1. this is now in m/s. TODO: cjheck this

        delta_lambda = (
            wavelengths * v_dopp / const_c
        )  # this only works if wavelengths in meters. TODO: check this

        # interpolate the spectrum onto the new wavelength grid
        spectrum_grid[i] = np.interp(wavelengths + delta_lambda, wavelengths, flux)
        # todo: make the array math line up. and be less confusing.

    # doppler shift based on the wavel
    return spectrum_grid


# @njit
def calc_planet_locations(phases, r_star, inc, lambda_misalign, a):
    """
    calculate the x and y positions of the planet on the stellar disk.

    the phase is at 0.0 when the planet is in front of the star. 0.5 when it's behind the star.


    phases are in units of orbital phase. not radians.

    a and rstar just need to be in the same units.

    inclination is in radians
    lambda_misalign is in radians.


    we need
    note: theta is a bit all over the place for the center of the star. but that's more or less physical : )

    Assumes a circular orbit, for now...
    :param phases:
    :param r_p:
    :param r_star:
    :param inc:
    :param lambda_misalign:
    :param kp:
    :return:
    """

    # todo: add the spatial extent of the planet. for now, just assume it's a point source.

    positions = np.zeros(
        (len(phases), 2)
    )  # r and theta positions! need to know the lambda misalignment though lol
    for i, phase in enumerate(phases):
        # calculate the x and y positions of the planet on the stellar disk
        x = a * np.sin(2 * np.pi * phase) / r_star
        y = np.cos(inc) * a * np.cos(2 * np.pi * phase) / r_star

        # calculate the r and theta positions
        r = np.sqrt(x**2 + y**2)
        theta = np.arctan2(y, x)

        # now rotate the theta position by the misalignment angle
        theta += lambda_misalign

        positions[i] = r, theta
    return positions


# @njit
def occult_grid(occulted_grid, grid, planet_location, r_p):
    """
    take the grid and the planet locations and occult the planet. grid is SPATIAL grid.

    r_p is given in stellar radii — so, it's in grid units.

    :param grid:
    :param planet_locations:
    :param r_p:
    :return:
    """
    # calculate the radial distance from the planet
    r1 = grid[:, 0]
    r2 = planet_location[0]
    theta1 = grid[:, 1]
    theta2 = planet_location[1]

    # calculate the distance between the planet and the grid points
    r = np.sqrt(r1**2 + r2**2 - 2 * r1 * r2 * np.cos(theta1 - theta2))

    # find the points where the planet is
    planet_points = np.where(r <= r_p)

    # set those points to zero
    occulted_grid[planet_points] = 0.0
    return occulted_grid


# @njit
def sum_grid(occulted_grid, areas):
    # todo: calculate the areas ahead of time. then just multiply by the spectrum.
    """
    sum up the grid. normalize by area.

    grid is the r theta situation
    :param grid:
    :return:
    """
    # pdb.set_trace()
    # what if i summed a slice
    summed_spectrum = np.dot((occulted_grid).T, areas) / np.sum(areas)
    # summed_spectrum = np.zeros(spectrum_grid.shape[1])
    # # can probably just multiply and broadcaset
    # for i, area in enumerate(areas):
    #     spectrum = spectrum_grid[i]
    #     summed_spectrum += area * spectrum

    # now it's in flux units that have area. it's ok if it's off by a multiplicative factor, I thnk. todo: check this.
    return summed_spectrum


def calc_areas(grid):
    """
    calculate the area of each cell in the grid.
    if it's an r, theta grid, then the area is r^2 * pi / (n_r * n_theta)
    no, that's not true. it involves dr. and dtheta. but we can just do it in r, theta space.

    so, the
    :param grid:
    :return:
    """
    areas = np.zeros(grid.shape[0])
    dr = grid[1, 0] - grid[0, 0]
    dtheta = np.max(np.diff(grid[:, 1]))
    # pdb.set_trace()
    for i, point in enumerate(grid):
        r, theta = point

        r_outer = r + dr / 2
        r_inner = r - dr / 2
        area = 0.5 * (r_outer**2 - r_inner**2) * dtheta
        areas[i] = area

    return areas


def make_stellar_disk(
    flux,
    wavelengths,
    v_rot,
    phases,
    r_star,
    inc,
    lambda_misalign,
    a,
    r_p,
    n_theta=10,
    n_r=10,
):
    """
    based on the x and y position of the planet on the disk, make the stellar spectrum.

    note: doesn't incorporate CLV variations if I just use a phoenix model.

    and if i use an empirical spectrum, I'm not blocking out CLV variations.

    take the average of all these approaches? lol

    v_rot is in km/s

    r_p in meters, and same with rstar and a
    """

    # step 1: make the stellar grid. it's in polar so we don't have to worry about edge clipping. careful about summing up the area, though.

    grid = make_grid(n_r, n_theta)

    # step 2: doppler shift onto a bunch of different mu values. ignores CLV variations.
    spectrum_grid = doppler_shift_grid(
        grid, flux, wavelengths, v_rot
    )  # spectrum as a function of r and theta

    # step 3: where is the planet on the grid?
    planet_locations = calc_planet_locations(
        phases, r_star, inc, lambda_misalign, a
    )  # r and theta positions of the planet as a function of exposure

    # loop over the next bit: occurs at every phase.
    summed_grids = np.zeros((len(phases), len(wavelengths)))

    areas = calc_areas(grid)

    for i, phase in enumerate(phases):
        occulted_grid = np.copy(spectrum_grid)

        planet_location = planet_locations[i]

        # step 4: occult those spots on the grid. oh. yeah it's a mask. so just multiply by spectrum grid!
        occulted_grid = occult_grid(occulted_grid, grid, planet_location, r_p / r_star)

        # step 5: add up the rest of the spectra, normalized by area!
        summed_grids[i] = sum_grid(occulted_grid, areas)

        # clean up memory. these should be pretty big arrays...

    # step 6: what's that overcorrection factor?
    correction_factor = (
        summed_grids / flux
    )  # this is what you get when you divide flux(in_transit) / flux(out).

    # you'll want to divide your correction factor by the data.
    # because wait, shouldn't the summed grids basically be on the order of the spectrum?
    return summed_grids, correction_factor, areas, occulted_grid

    # can do this convolution style assuming the same stellar line profile...
    # frame it as an area calculation. total minus the little bit you're
    # but are phoenix models already broadened? nope, they're not!



import astropy.units as u
import numpy as np

# refactor the below two functions to be a single function


def check_args(args, data, quantity):
    """
    Check that the arguments are in the data dictionary.

    Parameters
    ----------
    args : list
        List of arguments to check.
    data : dict
        Dictionary of data.

    """
    for arg in args:
        if arg not in data:
            raise ValueError(
                f"{arg} not in data dictionary! It is required to calculate {quantity}!"
            )


def calculate_derived_parameters(data):
    """
    Calculate derived parameters from the input data.

    """
    # first, equatorial rotational velocity.

    def calc_param_boilerplate(param, args, data, distance_unit):
        if np.isnan(data[param]):
            check_args(args, data, param)
            data[param] = calculate_velocity(
                *[data[arg] for arg in args], distance_unit=distance_unit
            )

    # Calculate the equatorial rotational velocity
    calc_param_boilerplate("v_rot", ["Rp", "P_rot"], data, u.R_jup)

    # calculate planetary orbital velocity
    calc_param_boilerplate("kp", ["a", "P_rot"], data, u.AU)

    return data


def calculate_velocity(distance, P, distance_unit=u.AU, time_unit=u.day):
    """
    Calculate a velocity over some period over some distance.
    """
    # Calculate the velocity
    velocity = 2 * np.pi * distance * distance_unit / (P * time_unit)

    return velocity.to(u.km / u.s).value


def convert_tdur_to_phase(tdur, period):
    """
    Convert the transit duration to a phase.

    Parameters
    ----------
    tdur : float
        The transit duration in hours.
    period : float
        The period of the planet in days.

    Returns
    -------
    float
        The phase of the transit duration.
    """
    return ((tdur * u.hour) / (period * u.day)).si.value



import pytest

from scope.emcee_fit_hires import *
from scope.tests.conftest import test_baseline_outouts, test_inputs

test_data_path = os.path.join(os.path.dirname(__file__), "../data")


@pytest.mark.parametrize(
    "values, output",
    [
        ([150, 0, 0], 0.0),
        ([2, 0, 0], -np.inf),
        ([150, -200, 0], -np.inf),
        ([150, 0, -100], -np.inf),
        ([150, -200, -100], -np.inf),
        ([2, 0, -100], -np.inf),
        ([150, -200, 0], -np.inf),
        ([2, -200, -100], -np.inf),
    ],
)
def test_prior(values, output):
    """
    Test the prior function.
    """
    Kp, Vsys, log_scale = values
    assert prior(values, best_kp=160.0) == output


def test_likelihood(test_baseline_outouts, test_inputs):
    (
        A_noplanet,
        flux_cube,
        flux_cube_nopca,
        just_tellurics,
        n_exposure,
        n_order,
        n_pixel,
    ) = test_baseline_outouts
    star = True
    Rp_solar = 0.1
    Rstar = 1.0
    phases = np.linspace(-0.01, 0.01, n_exposure)
    Fp_conv, Fstar_conv, wl_cube_model, wl_model = test_inputs
    best_kp = 192.06

    x_good = [best_kp, 0, 0]
    x_bad = [144, -50, -0.9]
    n_princ_comp = 4
    log_prob_good = log_prob(
        x_good,
        best_kp,
        wl_cube_model,
        Fp_conv,
        n_order,
        n_exposure,
        n_pixel,
        A_noplanet,
        star,
        n_princ_comp,
        flux_cube,
        wl_model,
        Fstar_conv,
        Rp_solar,
        Rstar,
        phases,
        do_pca,
    )

    log_prob_bad = log_prob(
        x_bad,
        best_kp,
        wl_cube_model,
        Fp_conv,
        n_order,
        n_exposure,
        n_pixel,
        A_noplanet,
        star,
        n_princ_comp,
        flux_cube,
        wl_model,
        Fstar_conv,
        Rp_solar,
        Rstar,
        phases,
        do_pca,
    )

    assert log_prob_good > log_prob_bad



from functools import wraps

import jax
import jax.numpy as jnp


def jit(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return jax.jit(f(*args, **kwargs))

    return wrapper


# @jit
def calc_ccf(model_flux, data_arr_slice, n_pixel):
    """
    Calculates the CCF between a model and a data slice.

    Inputs
    ------
        :model_flux: (jnp.ndarray) The model flux, normalized and such.
        :data_arr_slice: (jnp.ndarray) The data slice, normalized and such.
        :n_pixels: (int) The number of pixels in the data slice.

    Outputs
    -------
        :logl: (float) The log-likelihood of the data given the model.
        :CCF: (float) The CCF between the model and the data.

    """
    model_vector = jnp.subtract(
        model_flux, jnp.vstack(jnp.mean(model_flux, axis=1))
    )  # normalized and such
    variance_model = jnp.var(model_vector, axis=1)
    variance_data = jnp.var(data_arr_slice, axis=1)
    cross_variance = (model_vector * data_arr_slice).sum(axis=1) / n_pixel

    # now need to sum
    ccf = (cross_variance / jnp.sqrt(variance_data * variance_model)).sum()

    logl = (
        -0.5 * n_pixel * jnp.log(variance_data + variance_model - 2.0 * cross_variance)
    ).sum()

    return logl, ccf


calc_ccf_map = jax.vmap(calc_ccf, in_axes=(0, None, None), out_axes=0)



import os
import pickle

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
from exoplanet.orbits.keplerian import KeplerianOrbit
from numba import njit
from scipy.interpolate import splev, splrep
from scipy.ndimage import gaussian_filter1d
from tqdm import tqdm

from scope.constants import *

abs_path = os.path.dirname(__file__)

np.random.seed(42)
start_clip = 200
end_clip = 100


@njit
def doppler_shift_planet_star(
    model_flux_cube,
    n_exposure,
    phases,
    rv_planet,
    rv_star,
    wlgrid_order,
    wl_model,
    Fp_conv,
    Rp_solar,
    Fstar_conv,
    Rstar,
    u1,
    u2,
    a,
    b,
    LD,
    scale,
    star,
    observation,
    reprocessing=False,
):
    for exposure in range(n_exposure):
        flux_planet = calc_doppler_shift(
            wlgrid_order, wl_model, Fp_conv, rv_planet[exposure]
        )
        flux_planet *= scale  # apply scale factor
        flux_star = calc_doppler_shift(
            wlgrid_order, wl_model, Fstar_conv, rv_star[exposure]
        )

        if star and observation == "emission":
            model_flux_cube[exposure,] = (flux_planet * Rp_solar**2) / (
                flux_star * Rstar**2
            ) + 1.0
            if not reprocessing:
                model_flux_cube[exposure,] *= flux_star * Rstar**2

        else:  # in transmission, after we "divide out" (with PCA) the star and tellurics, we're left with Fp.
            I = calc_limb_darkening(u1, u2, a, b, Rstar, phases[exposure], LD)
            model_flux_cube[exposure,] = 1.0 - flux_planet * I
            if not reprocessing:
                model_flux_cube[exposure,] *= flux_star
    return model_flux_cube


def save_results(outdir, run_name, lls, ccfs):
    np.savetxt(
        f"{outdir}/lls_{run_name}.txt",
        lls,
    )
    np.savetxt(
        f"{outdir}/ccfs_{run_name}.txt",
        ccfs,
    )


def make_outdir(outdir):
    try:
        os.mkdir(outdir)
    except FileExistsError:
        print("Directory already exists. Continuing!")


def get_instrument_kernel(resolution_ratio=250000 / 45000):
    xker = np.arange(41) - 20
    sigma = resolution_ratio / (2.0 * np.sqrt(2.0 * np.log(2.0)))  # nominal
    yker = np.exp(-0.5 * (xker / sigma) ** 2.0)
    yker /= yker.sum()
    return yker


def save_data(outdir, run_name, flux_cube, flux_cube_nopca, A_noplanet, just_tellurics):
    """
    Saves data to a pickle file.

    Inputs
    ------
        :outdir: (str) output directory
        :run_name: (str) name of the run
        :data: (dict) data to save
    """
    with open(
        f"{outdir}/simdata_{run_name}.txt",
        "wb",
    ) as f:
        pickle.dump(flux_cube, f)
    with open(
        f"{outdir}/nopca_simdata_{run_name}.txt",
        "wb",
    ) as f:
        pickle.dump(flux_cube_nopca, f)
    with open(
        f"{outdir}/A_noplanet_{run_name}.txt",
        "wb",
    ) as f:
        pickle.dump(A_noplanet, f)

    with open(f"{outdir}/just_tellurics_vary_airmass.txt", "wb") as f:
        pickle.dump(just_tellurics, f)


@njit
def calc_limb_darkening(u1, u2, a, b, Rstar, ph, LD):
    """
    calculates limb darkening as a function of phase

    :u1: (float) linear limb darkening coefficient
    :u2: (float) quadratic limb darkening coefficient
    :a: (float) semi-major axis in meters
    :b: (float) impact parameter
    :Rstar: (float) stellar radius in solar radii
    :ph: (array) phase
    :LD: (bool) whether to apply limb darkening
    """
    if LD:  # apply limb darkening. 1D style!
        x = (a * np.sin(2 * np.pi * ph)) / (Rstar * rsun)
        mu = np.sqrt(1 - x**2 - b**2)
        if x**2 <= 1 - b**2:
            I = 1 - u1 * (1 - mu) - u2 * (1 - mu) ** 2
        else:
            I = 0.0
    else:
        I = 1.0
    return I


# todo: download atran scripts
# todo: fit wavelength solution stuff
# todo: plot the maps


@njit
def perform_pca(input_matrix, n_princ_comp, return_noplanet=False):
    """
    Perform PCA using SVD.

    SVD is written as A = USV^H, where ^H is the Hermitian operator.

    Inputs
    ------
        :input_matrix:
        :n_princ_comp: number of principal components to keep
    """
    u, singular_values, vh = np.linalg.svd(
        input_matrix, full_matrices=False
    )  # decompose
    if return_noplanet:
        s_high_variance = singular_values.copy()
        s_high_variance[n_princ_comp:] = 0.0  # keeping the high-variance terms here
        s_matrix = np.diag(s_high_variance)
        A_noplanet = np.dot(u, np.dot(s_matrix, vh))

    singular_values[:n_princ_comp] = 0.0  # zero out the high-variance terms here
    s_matrix = np.diag(singular_values)
    arr_planet = np.dot(u, np.dot(s_matrix, vh))

    if return_noplanet:
        return arr_planet, A_noplanet
    return arr_planet, arr_planet


@njit
def calc_doppler_shift(eval_wave, template_wave, template_flux, v):
    """
    Doppler shifts a spectrum. Evaluates the flux at a different grid.
    convention: positive v is redshift.

    Inputs
    ------
        :wl: wavelength grid
        :flux: flux grid
        :v: velocity. Must be in m/s.

    Outputs
    -------
        :flux_shifted: shifted flux grid
    """
    beta = v / const_c
    delta_lam = eval_wave * beta
    shifted_wave = eval_wave - delta_lam
    shifted_flux = np.interp(shifted_wave, template_wave, template_flux)
    return shifted_flux


def calc_crossing_time(
    period=1.80988198,
    mstar=1.458,
    e=0.000,
    inc=89.623,
    mplanet=0.894,
    rstar=1.756,
    peri=0,
    b=0.027,
    R=45000,
    pix_per_res=3.3,
    phase_start=0.9668567402328337,
    phase_end=1.0331432597671664,
    plot=False,
):
    """

    todo: refactor this into separate functions, maybe?

    Inputs
    -------
    autofilled for WASP-76b and IGRINS.

    R: (int) resolution of spectrograph.
    pix_per_res: (float) pixels per resolution element


    Outputs
    -------

    min_time: minimum time during transit before lines start smearing across resolution elements.

    min_time_per_pixel: minimum time during transit before lines start smearing across a single pixel.
    dphase_per_exp: the amount of phase (values from 0 to 1) taken up by a single exposure, given above constraints.
    n_exp: number of exposures you can take during transit.

    """

    orbit = KeplerianOrbit(
        m_star=mstar,  # solar masses!
        r_star=rstar,  # solar radii!
        #     m_planet_units = u.jupiterMass,
        t0=0,  # reference transit at 0!
        period=period,
        ecc=e,
        b=b,
        omega=np.radians(peri),  # periastron, radians
        Omega=np.radians(peri),  # periastron, radians
        m_planet=mplanet * 0.0009543,
    )
    t = np.linspace(0, period * 1.1, 1000)  # days
    z_vel = orbit.get_relative_velocity(t)[2].eval()

    z_vel *= 695700 / 86400  # km / s

    phases = t / period
    if plot:
        plt.plot(phases, z_vel)

        plt.axvline(0.5, color="gray", label="Eclipse")
        plt.axvline(0.25, label="Quadrature (should be maximized)", color="teal")

        plt.axvline(0.75, label="Quadrature (should be maximized)", color="teal")

        plt.legend()
        plt.xlabel("Time (days)")
        plt.ylabel("Radial velocity (km/s)")
    acceleration = (
        np.diff(z_vel) / np.diff((t * u.day).to(u.s).si.value) * u.km / (u.s**2)
    )

    if plot:
        plt.plot(phases[:-1], acceleration)

        plt.axvline(0.5, color="gray", label="Eclipse")
        plt.axvline(0.25, label="Quadrature (should be minimized)", color="teal")

        plt.axvline(0.75, label="Quadrature (should be minimized)", color="teal")
        plt.xlabel("Orbital phase")
        plt.ylabel("Radial acceleration (km/s^2)")

        # cool, this is the acceleration.

        # now want the pixel crossing time.

        plt.legend()

    # R = c / delta v
    # delta v = c / R

    delta_v = const.c / R
    delta_v = delta_v.to(u.km / u.s)
    res_crossing_time = abs(delta_v / acceleration).to(u.s)
    if plot:
        plt.figure()
        plt.plot(phases[:-1], res_crossing_time)
        plt.axvline(0.5, color="gray", label="Eclipse")
        plt.axvline(0.25, label="Quadrature (should be maximized)", color="teal")

        plt.axvline(0.75, label="Quadrature (should be maximized)", color="teal")

        plt.legend()
        plt.xlabel("Orbital phase")
        plt.ylabel("Resolution crossing time (s)")
        plt.yscale("log")

        plt.figure()
        plt.plot(phases[:-1], res_crossing_time)
        plt.legend()

        # plt.yscale('log')
        plt.ylim(820, 900)
        plt.xlabel("Orbital phase")
        plt.ylabel("Resolution crossing time (s)")

        plt.xlim(0.96, 1.041)

    # todo: generalize this!
    ingress = phase_start
    egress = phase_end
    during_transit = (phases[:-1] > ingress) & (phases[:-1] < egress)

    res_crossing_time_transit = res_crossing_time[during_transit]

    max_time = np.min(res_crossing_time_transit)

    max_time_per_pixel = max_time / pix_per_res
    period = period * u.day
    dphase_per_exp = (np.min(res_crossing_time_transit) / period).si

    transit_dur = (4.3336 / 24) / period.value  # degrees. todo: calculate.
    print(transit_dur)
    print(4.3336 * 60 * 60)  # this is how many seconds long it is
    n_exp = transit_dur / dphase_per_exp

    # then query https://igrins-jj.firebaseapp.com/etc/simple:
    # this gives us, for the given exposure time, what the SNR is going to be.
    # well that's more the maximum time
    # so that's the maximum time, but we want more than that many exposures.
    # don't have to worry about pixel-crossing time.
    return max_time, max_time_per_pixel, dphase_per_exp, n_exp


@njit
def calc_rvs(v_sys, v_sys_measured, Kp, Kstar, phases):
    """
    calculate radial velocities of planet and star.

    Inputs
    ------
    v_sys: float
        Systemic velocity of the system. Measured in km/s
    v_sys_measured: float
        Measured systemic velocity of the system. Measured in km/s
    Kp: float
        Planet semi-amplitude. Measured in km/s
    Kstar: float
        Star semi-amplitude. Measured in km/s
    phases: array
        Orbital phases of the system. Measured in radians.

    Returns
    -------
    rv_planet: array
        Radial velocities of the planet. Measured in m/s
    rv_star: array
        Radial velocities of the star. Measured in m/s

    """
    v_sys_tot = v_sys + v_sys_measured  # total offset
    rv_planet = (
        v_sys_tot + Kp * np.sin(2.0 * np.pi * phases) * 1e3
    )  # input in km/s, convert to m/s

    rv_star = (
        v_sys_measured - Kstar * np.sin(2.0 * np.pi * phases)
    ) * 1e3  # measured in m/s. note opposite sign!
    return rv_planet, rv_star


def get_star_spline(star_wave, star_flux, planet_wave, yker, smooth=True):
    """
    calculates the stellar spline. accounting for convolution and such.

    Inputs
    ------
    star_wave: array
        Wavelengths of the star. Measured in microns.
    star_flux: array
        Fluxes of the star. Measured in W/m^2/micron.
    planet_wave: array
        Wavelengths of the planet. Measured in microns.
    yker: array
        Convolution kernel.
    smooth: bool
        Whether or not to smooth the star spectrum. Default is True.

    Returns
    -------
    star_flux: array
        Fluxes of the star. Measured in W/m^2/micron. todo: check.
    """
    star_spline = splrep(star_wave, star_flux, s=0.0)

    star_flux = splev(
        planet_wave, star_spline, der=0
    )  # now on same wavelength grid as planet wave

    star_flux = np.convolve(star_flux, yker, mode="same")  # convolving star too

    if smooth:
        star_flux = gaussian_filter1d(star_flux, 200)

    return star_flux


def change_wavelength_solution(wl_cube_model, flux_cube_model, doppler_shifts):
    """
    Takes a finalized wavelength cube and makes the wavelength solution for each exposure just slightly wrong.

    for now, it just shifts things. doesn't stretch.

    this basically interpolates the flux array onto a different wavelength grid. don't worry about the edge, it'll be
    trimmed?

    for now: uses prescribed pixel shifts!

    Inputs
    ------
        :wl_cube_model: (array) wavelength grid of simulated data!
        :flux_cube_model: (array) flux cube for simulated data!
        :pixel_shifts: (array) N_exp long array of number of wavelength pixels to shift. Convention: positive number is
                        redshift (shift to larger pixels). shifts must be in km/s.
    """
    n_order = flux_cube_model.shape[0]

    # iterate through each exposure
    for exp, doppler_shift in enumerate(doppler_shifts):
        doppler_shift *= 1e3  # convert to m/s

        # iterate through each exposure
        for order in range(n_order):
            wl_grid = wl_cube_model[order]
            flux = flux_cube_model[order][exp]
            flux_cube_model[order][exp] = calc_doppler_shift(
                wl_grid, wl_grid, flux, doppler_shift
            )

    return flux_cube_model


def add_blaze_function(wl_cube_model, flux_cube_model, n_order, n_exp):
    """
    Adds the blaze function to the model.

    Inputs
    ------
        :wl_cube_model: (array) wavelength cube model
        :flux_cube_model: (array) flux cube model
        :n_order: (int) number of orders
        :n_exp: (int) number of exposures

    Outputs
    -------
        :flux_cube_model: (array) flux cube model with blaze function included.
    """
    # read in...have to somehow match the telluric spectra

    with open(abs_path + "/data/K_blaze_spectra.pic", "rb") as f:
        K_blaze_cube = pickle.load(f)

    with open(abs_path + "/data/H_blaze_spectra.pic", "rb") as f:
        H_blaze_cube = pickle.load(f)

    n_orders_k = K_blaze_cube.shape[0]
    n_orders_h = H_blaze_cube.shape[0]
    # K first, then H
    K_blaze_cube = detrend_cube(K_blaze_cube, n_orders_k, n_exp, blaze=True)
    H_blaze_cube = detrend_cube(H_blaze_cube, n_orders_h, n_exp, blaze=True)

    for order in tqdm(range(n_order), desc="adding blaze function"):
        flux_cube_model_slice = flux_cube_model[order, :, :]
        if order >= n_orders_k:
            blaze_cube = H_blaze_cube
            order_used = order - 19
        else:
            blaze_cube = K_blaze_cube
            order_used = order
        flux_cube_model_slice = flux_cube_model_slice * blaze_cube[order_used][100:-100]
        flux_cube_model[order, :, :] = flux_cube_model_slice
    return flux_cube_model


def detrend_cube(cube, n_order, n_exp, blaze=False):
    """
    Detrends the cube by dividing each order by its maximum value.

    Inputs
    ------
        :cube: (array) flux cube
        :n_order: (int) number of orders
        :n_exp: (int) number of exposures

    Outputs
    -------
        :cube: (array) detrended flux cube
    """

    def detrend_array(array):
        max_val = np.nanmax(array)
        return array / max_val

    progress = tqdm(range(n_order), desc="detrending cube")

    if not blaze:
        for order in progress:
            for exp in range(n_exp):
                cube[order, exp] = detrend_array(cube[order, exp])

    else:
        for order in progress:
            cube[order] = detrend_array(cube[order])

    return cube


def unpack_grid(grid_ind, parameter_list):
    param_dict = parameter_list[grid_ind]

    (
        blaze,
        n_princ_comp,
        star,
        SNR,
        telluric,
        tell_type,
        time_dep_tell,
        wav_error,
        order_dep_throughput,
    ) = (
        param_dict["blaze"],
        param_dict["n_princ_comp"],
        param_dict["star"],
        param_dict["SNR"],
        param_dict["telluric"],
        param_dict["telluric_type"],
        param_dict["time_dep_telluric"],
        param_dict["wav_error"],
        param_dict["order_dep_throughput"],
    )
    return (
        blaze,
        n_princ_comp,
        star,
        SNR,
        telluric,
        tell_type,
        time_dep_tell,
        wav_error,
        order_dep_throughput,
    )



import unittest

from scope.broadening import *


class TestTheta(unittest.TestCase):
    def test_get_theta_zero(self):
        res = get_theta(0, 0)
        self.assertTrue(res == 0)

    def test_get_theta_90_lat(self):
        res = get_theta(np.pi / 2, 0)
        self.assertTrue(res == np.pi / 2)

    def test_get_theta_90_lon(self):
        res = get_theta(0, np.pi / 2)
        self.assertTrue(res == np.pi / 2)

    def test_more_than_each(self):
        """
        at positive lat and lon, there should be more theta than either lat or lon
        """
        res = get_theta(np.pi / 4, np.pi / 4)
        self.assertTrue(res > np.pi / 4)

    def test_get_theta_lat_too_large_positive(self):
        """
        if we get a value that's greater than pi, we should yell.
        """
        self.assertRaises(ValueError, get_theta, 1.1 * np.pi, 0)

    def test_get_theta_lat_too_large_positive(self):
        """
        if we get a value that's greater than pi, we should yell.
        """
        self.assertRaises(
            ValueError,
            get_theta,
            0,
            1.1 * np.pi,
        )


class TestGaussian(unittest.TestCase):
    """
    need to test this! so bad!
    """

    def test_lat_lon_same(self):
        res1 = gaussian_term(0.3, 0, 0, 0.1, 1)
        res2 = gaussian_term(0, 0.3, 0, 0.1, 1)

        self.assertTrue(res1 == res2)

    def test_lat_lon_not_same_with_offset(self):
        res1 = gaussian_term(0.3, 0, 0.3, 0.1, 1)
        res2 = gaussian_term(0, 0.3, 0, 0.1, 1)

        self.assertTrue(res1 > res2)

    def test_broadening_lon(self):
        res1 = gaussian_term(0.3, 0, 0, 0.1, 1)
        res2 = gaussian_term(0.3, 0, 0, 0.2, 1)

        self.assertTrue(res2 > res1)

    def test_broadening_lat(self):
        res1 = gaussian_term(0, 0.3, 0, 0.1, 1)
        res2 = gaussian_term(0, 0.3, 0, 0.2, 1)

        self.assertTrue(res2 > res1)

    def test_broadening_both(self):
        res1 = gaussian_term(0.3, 0.3, 0, 0.1, 1)
        res2 = gaussian_term(0.3, 0.3, 0, 0.2, 1)

        self.assertTrue(res2 > res1)

    def test_amp(self):
        res1 = gaussian_term(0.3, 0.3, 0, 0.1, 1)
        res2 = gaussian_term(0.3, 0.3, 0, 0.2, 2)

        self.assertTrue(res2 > res1)


class TestI_darken(unittest.TestCase):
    """
    test all the functions that do the actual darkening
    """

    def test_I_darken_zero_epsilon(self):
        """
        if there's no epsilon, it's zero at the edge and
        """

        outside = I_darken(np.pi / 4, np.pi / 4, 0)
        center = I_darken(0, 0, 0)
        self.assertTrue(center == outside)

    def test_I_darken_pos_epsilon(self):
        """
        if there's a positive epsilon, it should be less at edge than center.
        """
        outside = I_darken(np.pi / 4, np.pi / 4, 0.3)
        center = I_darken(0, 0, 0.3)
        self.assertTrue(center > outside)

    def test_I_darken_only_pos_epsilon(self):
        """
        should throw an error if epsilon is negative.
        """

        self.assertRaises(ValueError, I_darken, 0, 0, -0.3)

    def test_I_darken_only_epsilon_less_one(self):
        """
        should throw an error if epsilon > 1
        """

        self.assertRaises(ValueError, I_darken, 0, 0, 1.3)

    def test_I_darken_disk_x_greater_one(self):
        """
        any x greater than 1 shouldn't be good.
        """
        self.assertTrue(I_darken_disk(1.5, 0, 0.1) == 0.0)

    def test_I_darken_disk_y_greater_one(self):
        """
        any x greater than 1 shouldn't be good.
        """

        self.assertTrue(I_darken_disk(0.0, 1.5, 0.1) == 0.0)

    def test_I_darken_disk_same_on_disk(self):
        """
        any x greater than 1 shouldn't be good.
        """
        res1 = I_darken_disk(0, 0.5, 0)
        res2 = I_darken_disk(0.3, 0.1, 0)

        self.assertTrue(res1 == res2)

    def test_I_darken_disk_disk_decrease(self):
        """
        any x greater than 1 shouldn't be good.
        """
        res1 = I_darken_disk(0, 0, 0.3)
        res2 = I_darken_disk(0.3, 0.1, 0.3)

        self.assertTrue(res1 > res2)

    def test_I_darken_disk_disk_decrease(self):
        """
        any x greater than 1 shouldn't be good.
        """
        res1 = I_darken_disk(0, 0, 0.3)
        res2 = I_darken_disk(0.3, 0.1, 0.3)

        self.assertTrue(res1 > res2)

        self.assertTrue(res2 < res1)

        # I_darken_hotspot_add(lon, lat, offset, sigma, epsilon, amp):


class Test_Profile(unittest.TestCase):
    def test_profile_offset_disk(self):
        """
        when it's offset, it should be more redshifted
        """

        self.assertTrue(True)
